"""Test the OUTCAR parser."""
# pylint: disable=unused-import,redefined-outer-name,unused-argument,unused-wildcard-import,wildcard-import
# pylint: disable=invalid-name

import numpy as np
import pytest


@pytest.mark.parametrize('outcar_parser', ['outcar'], indirect=True)
def test_parse_outcar_bogus_quantity(outcar_parser):
    """Load a reference OUTCAR parser.

    We check that the get quantity returns None for bogus keys.

    """

    assert outcar_parser.get_quantity('bogusquantity1234') is None


@pytest.mark.parametrize('outcar_parser', ['outcar'], indirect=True)
def test_parse_outcar_no_quantity(outcar_parser):
    """Load a reference OUTCAR parser.

    We check that the get quantity returns None if that quantity is not found
    in the presented OUTCAR.

    """
    # A standard OUTCAR should not contain parsable magnetization.
    assert outcar_parser.get_quantity('magnetization') is None


@pytest.mark.parametrize(
    ['outcar_parser'],
    [(['disp_details', 'OUTCAR', {'quantities_to_parse': ['symmetries']}],)],
    indirect=True,
)
def test_parse_outcar_symmetry(outcar_parser, compare_symmetries):
    """Load a reference OUTCAR parser.

    We check that it parses and provides the correct content for the symmetry.

    """

    symmetry = outcar_parser.get_quantity('symmetries')
    assert set(symmetry) == set(compare_symmetries)


@pytest.mark.parametrize(
    ['outcar_parser'],
    [(['disp_details', 'OUTCAR', {'quantities_to_parse': ['elastic_moduli']}],)],
    indirect=True,
)
def test_parse_outcar_elastic_moduli(outcar_parser):
    """Load a reference OUTCAR parser.

    We check that it parses and provides the correct content for the elastic moduli.

    """

    moduli = outcar_parser.get_quantity('elastic_moduli')
    test = np.array([1674.5786, 704.739, 704.739, -0.0, 0.0, 0.0])
    np.testing.assert_allclose(moduli['symmetrized'][0], test)
    test = np.array([0.0, 0.0, 0.0, -0.0, -0.0, 1122.6622])
    np.testing.assert_allclose(moduli['symmetrized'][5], test)
    test = np.array([705.0238, 1674.8491, 705.0238, -0.0, -0.0, 0.0])
    np.testing.assert_allclose(moduli['non_symmetrized'][1], test)
    test = np.array([-0.0078, -0.0495, 0.0147, 0.0, 1123.0829, -0.0])
    np.testing.assert_allclose(moduli['non_symmetrized'][4], test)
    test = np.array([704.739, 704.739, 1674.5786, -0.0, -0.0, 0.0])
    np.testing.assert_allclose(moduli['total'][2], test)
    test = np.array([-0.0, -0.0, -0.0, 775.8054, 0.0, -0.0])
    np.testing.assert_allclose(moduli['total'][3], test)


@pytest.mark.parametrize('outcar_parser', ['disp_details'], indirect=True)
def test_parse_outcar_status(outcar_parser):
    """Load a reference OUTCAR parser.

    We check that it parses and provides the correct content for a simple status.

    """

    status = outcar_parser.get_quantity('run_status')
    assert status['last_iteration_index'] == [15, 5]
    assert status['finished']
    assert status['ionic_converged']
    assert status['electronic_converged']
    assert status['nelm'] == 60
    assert status['nsw'] == 61


_TEST_DATA = [
    (['outcar_extras', 'OUTCAR.converged'], [True, True, True, False, False]),
    (
        ['outcar_extras', 'OUTCAR.nelm-breach-consistent'],
        [True, False, False, True, True],
    ),
    (['outcar_extras', 'OUTCAR.nelm-breach-partial'], [True, False, True, False, True]),
    (['outcar_extras', 'OUTCAR.unfinished'], [False, False, False, False, False]),
    (['outcar_extras', 'OUTCAR.not-converged'], [True, False, True, False, False]),
]


@pytest.mark.parametrize('outcar_parser,expected', _TEST_DATA, indirect=['outcar_parser'])
def test_parse_outcar_status_extended(outcar_parser, expected):
    """Load a reference OUTCAR parser.

    We check that it parses and provides the correct content for the status for multiple OUTCARs.

    """

    status = outcar_parser.get_quantity('run_status')
    assert status['finished'] is expected[0]
    assert status['ionic_converged'] is expected[1]
    assert status['electronic_converged'] is expected[2]
    assert status['consistent_nelm_breach'] is expected[3]
    assert status['contains_nelm_breach'] is expected[4]


@pytest.mark.parametrize(
    ['outcar_parser'],
    [
        (
            [
                'magnetization',
                'OUTCAR',
                {'quantities_to_parse': ['magnetization', 'site_magnetization']},
            ],
        )
    ],
    indirect=True,
)
def test_parse_outcar_magnetizationr(aiida_profile, outcar_parser):
    """Load a reference OUTCAR parser.

    We check that it parses and provides the correct content for the magnetization.

    """
    magnetization = outcar_parser.get_quantity('site_magnetization')
    test = {
        'sphere': {
            'x': {
                'site_moment': {
                    1: {'s': -0.014, 'p': -0.051, 'd': 1.687, 'tot': 1.621},
                    2: {'s': -0.015, 'p': -0.052, 'd': 1.686, 'tot': 1.619},
                    3: {'s': -0.014, 'p': -0.053, 'd': 1.708, 'tot': 1.64},
                    4: {'s': -0.014, 'p': -0.053, 'd': 1.708, 'tot': 1.64},
                },
                'total_magnetization': {
                    's': -0.057,
                    'p': -0.21,
                    'd': 6.788,
                    'tot': 6.521,
                },
            },
            'y': {'site_moment': {}, 'total_magnetization': {}},
            'z': {'site_moment': {}, 'total_magnetization': {}},
        },
        'full_cell': np.asarray([6.4424922]),
    }
    assert set(magnetization) == set(test)


@pytest.mark.parametrize(
    ['outcar_parser'],
    [
        (
            [
                'magnetization_single',
                'OUTCAR',
                {'quantities_to_parse': ['magnetization', 'site_magnetization']},
            ],
        )
    ],
    indirect=True,
)
def test_parse_outcar_magnetization_single(aiida_profile, outcar_parser):  # pylint: disable=invalid-name
    """Load a reference OUTCAR parser.

    We check that it parses and provides the correct content for the magnetization
    for a single atom in the cell.

    """
    magnetization = outcar_parser.get_quantity('site_magnetization')
    test = {
        'sphere': {
            'x': {
                'site_moment': {
                    1: {'s': -0.012, 'p': -0.043, 'd': 2.49, 'tot': 2.434},
                },
                'total_magnetization': {
                    's': -0.012,
                    'p': -0.043,
                    'd': 2.49,
                    'tot': 2.434,
                },
            },
            'y': {'site_moment': {}, 'total_magnetization': {}},
            'z': {'site_moment': {}, 'total_magnetization': {}},
        },
        'full_cell': np.asarray([2.4077611]),
    }
    assert set(magnetization) == set(test)


@pytest.mark.parametrize('neb_outcar_parser', ['neb/01'], indirect=True)
def test_neb(aiida_profile, neb_outcar_parser):
    """
    Test that the parameter node is a ParametersData instance.

    Should contain the symmetries and the elastic moduli.

    """
    data = neb_outcar_parser.get_quantity('neb_data')
    assert data['neb_converged']
    assert data['force_prep_real'] == pytest.approx(0.017467)
    assert neb_outcar_parser.total_energies['energy_extrapolated'] == pytest.approx(-19.49550593)

    np.testing.assert_allclose(neb_outcar_parser.forces[0], np.array([0.008815, 0.005492, -0.000661]))



"""
Wrapper class to simply updating an Node when configuring inputs of a calculation.

This module is currently not used in the code, but kept here for future reference.
It is not clear such wrapper for Dict would truly benefit the usability of the code.
"""

from collections import UserDict
from typing import Any

from aiida.orm.nodes.data.dict import Dict


class DictWrapper(UserDict):
    """
    Wrapper class for a dict

    Wraps around an Dict node, saved or not, to allow python dict-like in place modifications.
    """

    def __init__(self, node, namespace=None, port=None):
        """
        Wrapper around a ``Dict`` node. Optionally also include a namespace and a port that the
        underlying Node should be assigned to.
        The wrapped ``Dict`` node is never update in practice and only used as an initial reference data.
        If there is no change, the same Dict node will be passed as it is under the ``node`` property.

        :param node: The node to be wrapped.
        :param namespace: The namespace that the node should be assigned to.
        :param port: The port under the namespace.

        A new node is created whenever necessary.
        """
        super().__init__()

        self._stored_node = node
        self._unstored_node = None  # A new node to be used
        self.namespace = namespace
        self.port = port
        self.data = node.get_dict()
        if namespace is not None:
            namespace[port] = node

    def __getattr__(self, name: str) -> Any:
        return getattr(self.node, name)

    @property
    def is_updated(self):
        return self.data != self._stored_node.get_dict()

    @property
    def node(self):
        """
        Returns a node object that represents the stored data.

        This can be a stored node (if no update) or an unstore node
        """
        if self.is_updated:
            if self._unstored_node is None:
                self._unstored_node = Dict(dict=self.data)
            return self._unstored_node
        return self._stored_node

    def __setitem__(self, key, value):
        """
        Set the value of a key
        """
        update = False
        if key in self.data:
            if self.data[key] != value:
                update = True
        else:
            update = True
        self._ensure_unstored()
        super().__setitem__(key, value)
        # If we have updated the value of a key
        if update:
            if self._unstored_node is None:
                self._unstored_node = Dict(dict=self.data)
            else:
                self._unstored_node.base.attributes.set(key, value)

        if self.namespace is not None:
            self.namespace[self.port] = self.node

    def _ensure_unstored(self) -> None:
        """
        Ensure that self._unstored_node is indeed unstored
        If it is not the case - we move it to self._stored_node an create a new
        unstored node.
        """
        if self._unstored_node and self._unstored_node.is_stored:
            self._stored_node = self._unstored_node
            self._unstored_node = Dict(dict=self.data)

    def __delitem__(self, key) -> None:
        """
        Delete an item

        Delete from both the python dictionary and the underlying node
        """

        self._ensure_unstored()
        super().__delitem__(key)
        if self._unstored_node is None:
            self._unstored_node = Dict(dict=self.data)
        else:
            self._unstored_node.delete_attribute(key)

        if self.namespace is not None:
            self.namespace[self.port] = self.node

    def validate(self) -> None:
        """Validate the consistency between the node and the python dictionary"""
        assert self.data == self.node.get_dict()

    @classmethod
    def serializer(cls, data: dict | 'DictWrapper', port: Any = None) -> Dict:
        """
        The `serializer` function can be used for `ProcessSpec.input`.

        Returns a Dict node for the given data.
        """
        _ = port
        if isinstance(data, DictWrapper):
            return data.node
        return Dict(dict=data)



"""Test the KPOINTS parser."""
# pylint: disable=unused-import,redefined-outer-name,unused-argument,unused-wildcard-import,wildcard-import

import numpy as np
import pytest

from aiida_vasp.parsers.content_parsers.kpoints import KpointsParser


@pytest.mark.parametrize(['kpoints_parser'], [('kpoints',)], indirect=True)
def test_parse_kpoints_explicit(kpoints_parser):
    """Load a reference KPOINTS parser.

    We check that it parses and provides the correct content for mesh mode.

    """

    # The kpoints for KPOINTS parser should have the key `kpoints-kpoints`
    result = kpoints_parser.get_quantity('kpoints-kpoints')
    result_ref = {
        'comment': 'K-points',
        'divisions': None,
        'shifts': None,
        'points': [np.array([0.0, 0.0, 0.0]), np.array([0.0, 0.0, 0.5])],
        'weights': [1.0, 1.0],
        'tetra': None,
        'tetra_volume': None,
        'mode': 'explicit',
        'centering': None,
        'num_kpoints': 2,
        'cartesian': False,
        'generating_vectors': None,
    }
    compare_kpoints_content(result, result_ref)


@pytest.mark.parametrize(['kpoints_parser'], [(['kpoints', 'KPOINTS_mesh'],)], indirect=True)
def test_parse_kpoints_mesh(kpoints_parser):
    """Load a reference KPOINTS parser.

    We check that it parses and provides the correct content for mesh mode.

    """

    # The kpoints for KPOINTS parser should have the key `kpoints-kpoints`
    result = kpoints_parser.get_quantity('kpoints-kpoints')
    result_ref = {
        'comment': 'K-points',
        'divisions': [2, 2, 2],
        'shifts': [0.0, 0.0, 0.0],
        'points': None,
        'weights': None,
        'tetra': None,
        'tetra_volume': None,
        'mode': 'automatic',
        'centering': 'Monkhorst-Pack',
        'num_kpoints': 0,
        'cartesian': None,
        'generating_vectors': None,
    }
    compare_kpoints_content(result, result_ref)


@pytest.mark.parametrize(['kpoints_parser'], [('kpoints',)], indirect=True)
def test_parse_kpoints_write(kpoints_parser, tmpdir):
    """Load a reference KPOINTS parser and check that the write functionality works.

    Here we make sure the write function of the content parser works.

    """

    # Write the loaded reference KPOINTS content to file
    temp_path = str(tmpdir.join('KPOINTS'))
    kpoints_parser.write(temp_path)
    # Load the written KPOINTS using a new content parser instance and compare
    content = None
    with open(temp_path, 'r', encoding='utf8') as handler:
        content = handler.readlines()
    ref_content = [
        '# K-points\n',
        '     2\n',
        'Direct\n',
        '  0.000000000   0.000000000   0.000000000   1.000000000\n',
        '  0.000000000   0.000000000   0.500000000   1.000000000\n',
    ]
    assert content == ref_content


def test_parse_kpoints_data(vasp_kpoints, tmpdir):
    """Load a reference AiiDA KpointsData and check that the parser can
    initialize using the data.

    Using the KpointsData sitting in the initialized parser it should
    write that content to a KPOINTS file when issuing write which is also tested,
    file is reloaded and content checked.

    """

    # Initialize parser with an existing reference KpointsData
    kpoints_data, _ = vasp_kpoints
    kpoints_parser = KpointsParser(data=kpoints_data)

    # Check that get_quantity return the same KpointsData instance
    assert kpoints_data == kpoints_parser.get_quantity('key_does_not_matter')

    # Write the loaded KpointsData to file, which behind the scenes convert it
    # to a KPOINTS format
    temp_path = str(tmpdir.join('KPOINTS'))
    kpoints_parser.write(temp_path)

    # Load the written KPOINTS using a new content parser instance and compare
    parser = None
    with open(temp_path, 'r', encoding='utf8') as handler:
        parser = KpointsParser(handler=handler)
    result = parser.get_quantity('kpoints-kpoints')
    if result['mode'] == 'automatic':
        result_ref = {
            'comment': 'No comment',
            'divisions': [2, 2, 2],
            'shifts': [0.0, 0.0, 0.0],
            'points': None,
            'weights': None,
            'tetra': None,
            'tetra_volume': None,
            'mode': 'automatic',
            'centering': 'Gamma',
            'num_kpoints': 0,
            'cartesian': None,
            'generating_vectors': None,
        }
    else:
        result_ref = {
            'comment': 'No comment',
            'divisions': None,
            'shifts': None,
            'points': [np.array([0.0, 0.0, 0.0]), np.array([0.0, 0.0, 0.5])],
            'weights': [1.0, 1.0],
            'tetra': None,
            'tetra_volume': None,
            'mode': 'explicit',
            'centering': None,
            'num_kpoints': 2,
            'cartesian': False,
            'generating_vectors': None,
        }
    compare_kpoints_content(result, result_ref)


def compare_kpoints_content(kpoints, kpoints_ref):
    """Compare the KPOINTS content with supplied reference data."""
    if kpoints['mode'] == 'explicit':
        # Check all but points and weights
        kpoints_ref_no_points_weights = dict(kpoints_ref)
        del kpoints_ref_no_points_weights['points']
        del kpoints_ref_no_points_weights['weights']
        kpoints_no_points_weights = dict(kpoints)
        del kpoints_no_points_weights['points']
        del kpoints_no_points_weights['weights']
        assert kpoints_no_points_weights == kpoints_ref_no_points_weights
        # Check points and weights
        for index, item in enumerate(kpoints['points']):
            assert np.allclose(item, kpoints_ref['points'][index])
        assert np.allclose(kpoints['weights'], kpoints_ref['weights'])
    elif kpoints['mode'] == 'automatic':
        assert kpoints == kpoints_ref



import pytest
from aiida import orm
from aiida.common.links import LinkType
from aiida.orm import CalcJobNode
from aiida.plugins import DataFactory

from aiida_vasp.parsers.content_parsers.chgcar import ChgcarParser
from aiida_vasp.parsers.content_parsers.doscar import DoscarParser
from aiida_vasp.parsers.content_parsers.eigenval import EigenvalParser
from aiida_vasp.parsers.content_parsers.incar import IncarParser
from aiida_vasp.parsers.content_parsers.kpoints import KpointsParser
from aiida_vasp.parsers.content_parsers.outcar import OutcarParser, VtstNebOutcarParser
from aiida_vasp.parsers.content_parsers.poscar import PoscarParser
from aiida_vasp.parsers.content_parsers.stream import StreamParser
from aiida_vasp.parsers.content_parsers.vasprun import VasprunParser


@pytest.fixture()
def calc_with_retrieved(localhost):
    """A rigged CalcJobNode for testing the parser and that the calculation retrieve what is expected."""

    def _inner(file_path, input_settings=None):
        # Create a test computer
        computer = localhost

        process_type = 'aiida.calculations:vasp.vasp'

        node = CalcJobNode(computer=computer, process_type=process_type)
        node.base.attributes.set('input_filename', 'INCAR')
        node.base.attributes.set('output_filename', 'OUTCAR')
        # node.set_attribute('error_filename', 'aiida.err')
        node.base.attributes.set('scheduler_stderr', '_scheduler-stderr.txt')
        node.base.attributes.set('scheduler_stdout', '_scheduler-stdout.txt')
        node.set_option('resources', {'num_machines': 1, 'num_mpiprocs_per_machine': 1})
        node.set_option('max_wallclock_seconds', 1800)

        if input_settings is None:
            input_settings = {}

        settings = DataFactory('core.dict')(dict=input_settings)
        node.base.links.add_incoming(settings, link_type=LinkType.INPUT_CALC, link_label='settings')
        settings.store()
        node.store()

        # Create a `FolderData` that will represent the `retrieved` folder. Store the test
        # output fixture in there and link it.
        retrieved = DataFactory('core.folder')()
        retrieved.base.repository.put_object_from_tree(file_path)
        retrieved.base.links.add_incoming(node, link_type=LinkType.CREATE, link_label='retrieved')
        retrieved.store()

        return node

    return _inner


@pytest.fixture()
def vasp_structure_poscar(vasp_structure):
    """Fixture: Well formed POSCAR contents."""
    aiida_structure = vasp_structure
    if isinstance(vasp_structure, orm.CifData):
        ase_structure = vasp_structure.get_ase()
        aiida_structure = orm.StructureData(ase=ase_structure)
    writer = PoscarParser(data=aiida_structure)
    return writer


@pytest.fixture()
def vasprun_parser(request, data_path):
    """Return an instance of VasprunParser for a reference vasprun.xml. Remember rb mode."""
    path, settings = path_file_and_settings('vasprun.xml', request.param, data_path)
    with open(path, 'rb') as handler:
        parser = VasprunParser(handler=handler, settings=settings)
    return parser


@pytest.fixture()
def vasprun_parser_v621(request, data_path):
    """Return an instance of VasprunParser for a reference vasprun.xml of VASP6."""
    path, settings = path_file_and_settings('vasprun621.xml', request.param, data_path)
    with open(path, 'r', encoding='utf8') as handler:
        parser = VasprunParser(handler=handler, settings=settings)
    return parser


@pytest.fixture()
def neb_outcar_parser(request, data_path):
    """Return an instance of OutcarParser for a reference OUTCAR."""
    path, settings = path_file_and_settings('OUTCAR', request.param, data_path)
    with open(path, 'r', encoding='utf8') as handler:
        parser = VtstNebOutcarParser(handler=handler, settings=settings)
    return parser


@pytest.fixture()
def outcar_parser(request, data_path):
    """Return an instance of OutcarParser for a reference OUTCAR."""
    path, settings = path_file_and_settings('OUTCAR', request.param, data_path)
    with open(path, 'r', encoding='utf8') as handler:
        parser = OutcarParser(handler=handler, settings=settings)
    return parser


@pytest.fixture()
def poscar_parser(request, data_path):
    """Return an instance of PoscarParser for a reference POSCAR."""
    path, _ = path_file_and_settings('POSCAR', request.param, data_path)
    with open(path, 'r', encoding='utf8') as handler:
        parser = PoscarParser(handler=handler)
    return parser


@pytest.fixture()
def incar_parser(request, data_path):
    """Return an instance of IncarParser for a reference INCAR."""
    path, settings = path_file_and_settings('INCAR', request.param, data_path)
    with open(path, 'r', encoding='utf8') as handler:
        parser = IncarParser(handler=handler, settings=settings)
    return parser


@pytest.fixture()
def doscar_parser(request, data_path):
    """Return an instance of DoscarParser for a reference DOSCAR."""
    path, settings = path_file_and_settings('DOSCAR', request.param, data_path)
    with open(path, 'r', encoding='utf8') as handler:
        parser = DoscarParser(handler=handler, settings=settings)
    return parser


@pytest.fixture()
def chgcar_parser(request, data_path):
    """Return an instance of ChgcarParser for a reference CHGCAR."""
    path, settings = path_file_and_settings('CHGCAR', request.param, data_path)
    with open(path, 'r', encoding='utf8') as handler:
        parser = ChgcarParser(handler=handler, settings=settings)
    return parser


@pytest.fixture()
def eigenval_parser(request, data_path):
    """Return an instance of EigenvalParser for a reference EIGENVAL."""
    path, settings = path_file_and_settings('EIGENVAL', request.param, data_path)
    with open(path, 'r', encoding='utf8') as handler:
        parser = EigenvalParser(handler=handler, settings=settings)
    return parser


@pytest.fixture()
def kpoints_parser(request, data_path):
    """Return an instance of KpointsParser for a reference KPOINTS."""
    path, settings = path_file_and_settings('KPOINTS', request.param, data_path)
    with open(path, 'r', encoding='utf8') as handler:
        parser = KpointsParser(handler=handler, settings=settings)
    return parser


@pytest.fixture()
def stream_parser(request, data_path):
    """Return an instance of StreamParser for a reference stream capture."""
    path, settings = path_file_and_settings('vasp_output', request.param, data_path)
    with open(path, 'r', encoding='utf8') as handler:
        parser = StreamParser(handler=handler, settings=settings)
    return parser


def path_file_and_settings(name, param, data_path):
    """Locate folder, filename and settings from param. Return the path and settings."""
    settings = {}
    if isinstance(param, list):
        if len(param) == 3:
            folder, name, settings = param
        elif len(param) == 2:
            folder, name = param
        else:
            raise IndexError(
                'Please supply either folder and name, or folder, name and settings to the parser fixtures'
            )
    else:
        folder = param
    path = data_path(folder, name)

    return path, settings


@pytest.fixture
def compare_symmetries():
    return {
        'symmetrized_cell_type': {
            'static': [
                'face centered cubic supercell.',
                'body centered tetragonal supercell.',
                'body centered tetragonal supercell.',
                'body centered tetragonal supercell.',
                'body centered tetragonal supercell.',
                'body centered tetragonal supercell.',
                'body centered tetragonal supercell.',
                'base centered monoclinic supercell.',
                'base centered monoclinic supercell.',
                'base centered monoclinic supercell.',
                'base centered monoclinic supercell.',
                'base centered monoclinic supercell.',
                'base centered monoclinic supercell.',
                'face centered cubic supercell.',
                'face centered cubic supercell.',
                'face centered cubic supercell.',
            ],
            'dynamic': [
                'face centered cubic supercell.',
                'body centered tetragonal supercell.',
                'body centered tetragonal supercell.',
                'body centered tetragonal supercell.',
                'body centered tetragonal supercell.',
                'body centered tetragonal supercell.',
                'body centered tetragonal supercell.',
                'base centered monoclinic supercell.',
                'base centered monoclinic supercell.',
                'base centered monoclinic supercell.',
                'base centered monoclinic supercell.',
                'base centered monoclinic supercell.',
                'base centered monoclinic supercell.',
                'face centered cubic supercell.',
                'face centered cubic supercell.',
                'face centered cubic supercell.',
            ],
        },
        'original_cell_type': {
            'static': [
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
            ],
            'dynamic': [
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
                'primitive cell',
            ],
        },
        'num_space_group_operations': {
            'static': [48, 16, 16, 16, 16, 16, 16, 4, 4, 4, 4, 4, 4, 8, 8, 48],
            'dynamic': [48, 16, 16, 16, 16, 16, 16, 4, 4, 4, 4, 4, 4, 8, 8, 48],
        },
    }


@pytest.fixture()
def neb_calc_with_retrieved(localhost):
    """A rigged CalcJobNode for testing the parser and that the calculation retrieve what is expected."""

    def _inner(file_path, input_settings=None, nimgs=3):
        # Create a test computer
        computer = localhost

        process_type = 'aiida.calculations:vasp.vasp'

        node = CalcJobNode(computer=computer, process_type=process_type)
        node.base.attributes.set('input_filename', 'INCAR')
        node.base.attributes.set('output_filename', 'OUTCAR')
        # node.set_attribute('error_filename', 'aiida.err')
        node.base.attributes.set('scheduler_stderr', '_scheduler-stderr.txt')
        node.base.attributes.set('scheduler_stdout', '_scheduler-stdout.txt')
        node.set_option('resources', {'num_machines': 1, 'num_mpiprocs_per_machine': 1})
        node.set_option('max_wallclock_seconds', 1800)

        if input_settings is None:
            input_settings = {}

        settings = DataFactory('core.dict')(dict=input_settings)
        node.base.links.add_incoming(settings, link_type=LinkType.INPUT_CALC, link_label='settings')
        settings.store()

        # Add inputs with the number of images
        param = orm.Dict(dict={'images': nimgs})
        node.base.links.add_incoming(param, link_type=LinkType.INPUT_CALC, link_label='parameters')
        param.store()

        node.store()

        # Create a `FolderData` that will represent the `retrieved` folder. Store the test
        # output fixture in there and link it.
        retrieved = DataFactory('core.folder')()
        retrieved.base.repository.put_object_from_tree(file_path)
        retrieved.base.links.add_incoming(node, link_type=LinkType.CREATE, link_label='retrieved')
        retrieved.store()

        return node

    return _inner



# pylint: disable=too-many-function-args
"""
Mock vasp command.

Separate cli interface for commands useful in development and testing.
"""

import os
import pathlib
import shutil
from typing import List

import click

from aiida_vasp.parsers.content_parsers.incar import IncarParser
from aiida_vasp.parsers.content_parsers.kpoints import KpointsParser
from aiida_vasp.parsers.content_parsers.poscar import PoscarParser
from aiida_vasp.utils.mock_code import MockVasp, VaspMockRegistry, data_path


@click.command('mock-vasp-loose')
def mock_vasp_loose() -> None:
    """
    Loose version of mock-vasp that has default test data - only useful for testing and development.
    """
    return _mock_vasp(False)


@click.command('mock-vasp')
def mock_vasp() -> None:
    """
    If `MOCK_VASP_VASP_CMD` is set in the environment, it will use that command to run VASP if needed and add the
    calculation to the registry.
    """
    return _mock_vasp(True)


def _mock_vasp(strict_match: bool) -> None:  # pylint: disable=too-many-statements, too-many-locals, too-many-branches
    """
    Verify input objects are parsable and copy in output objects.
    """
    pwd = pathlib.Path().absolute()
    vasp_mock_output = []
    vasp_output_file = pwd / 'vasp_output'
    vasp_mock_output.append('MOCK PREPEND: START ----------------------\n')
    vasp_mock_output.append('MOCK PREPEND: Mock directory: ' + str(pwd) + '\n')

    incar = pwd / 'INCAR'
    if not incar.is_file():
        vasp_mock_output.append('MOCK PREPEND: INCAR not found.\n')
        stop_and_return(vasp_mock_output)

    potcar = pwd / 'POTCAR'
    if not potcar.is_file():
        vasp_mock_output.append('MOCK PREPEND: POTCAR not found.\n')
        stop_and_return(vasp_mock_output)

    poscar = pwd / 'POSCAR'
    if not poscar.is_file():
        vasp_mock_output.append('MOCK PREPEND: POSCAR not found.\n')
        stop_and_return(vasp_mock_output)

    kpoints = pwd / 'KPOINTS'
    if not kpoints.is_file():
        vasp_mock_output.append('MOCK PREPEND: KPOINTS not found.\n')
        stop_and_return(vasp_mock_output)

    # Check that the input files can be parsed (as close to a validity check we can get)
    incar_parser = False
    system = ''
    with open(str(incar), 'r', encoding='utf8') as handler:
        incar_parser = IncarParser(handler=handler, validate_tags=False)
        system = incar_parser.incar.get('system', '')
    if not incar_parser:
        vasp_mock_output.append('MOCK PREPEND: INCAR could not be parsed.\n')
        stop_and_return(vasp_mock_output)

    poscar_parser = False
    with open(str(poscar), 'r', encoding='utf8') as handler:
        poscar_parser = PoscarParser(handler=handler)
    if not poscar_parser:
        vasp_mock_output.append('MOCK PREPEND: POSCAR could not be parsed.\n')
        stop_and_return(vasp_mock_output)

    kpoints_parser = False
    with open(str(kpoints), 'r', encoding='utf8') as handler:
        kpoints_parser = KpointsParser(handler=handler)
    if not kpoints_parser:
        vasp_mock_output.append('MOCK PREPEND: KPOINTS could not be parsed.\n')
        stop_and_return(vasp_mock_output)

    try:
        test_case = system.strip().split(':')[1].strip()
    except IndexError:
        test_case = ''

    if not test_case:
        vasp_mock_output.append('MOCK PREPEND: Trying to detect test case using registry or reverting to default.\n')
        # If no test case is defined, we first try the hash-based mock registry
        mock_registry_path = os.environ.get('MOCK_VASP_REG_BASE', data_path('.'))
        mock_registry = VaspMockRegistry(mock_registry_path)
        vasp_mock_output.append(f'MOCK PREPEND: registry search paths: {mock_registry.search_paths}\n')

        # Setup the mock code
        mock = MockVasp(pwd, mock_registry, vasp_cmd=os.environ.get('MOCK_VASP_VASP_CMD'))
        if mock.is_runnable:
            mock.run()
            detected_path = mock.registry.get_path_by_hash(mock_registry.compute_hash(pwd))
            vasp_mock_output.append(
                f'MOCK PREPEND: Using test data in path {detected_path} based detection from inputs.\n'
            )
        else:
            vasp_mock_output.append(
                'MOCK PREPEND: Using default test data in the respective folders named similar to the file name.\n'
            )
            if not strict_match:
                # Then this is a simple case - assemble the outputs from folders
                shutil.copy(data_path('outcar', 'OUTCAR'), pwd / 'OUTCAR')
                shutil.copy(data_path('vasprun', 'vasprun.xml'), pwd / 'vasprun.xml')
                shutil.copy(data_path('chgcar', 'CHGCAR'), pwd / 'CHGCAR')
                shutil.copy(data_path('wavecar', 'WAVECAR'), pwd / 'WAVECAR')
                shutil.copy(data_path('eigenval', 'EIGENVAL'), pwd / 'EIGENVAL')
                shutil.copy(data_path('doscar', 'DOSCAR'), pwd / 'DOSCAR')
                shutil.copy(data_path('basic_run', 'vasp_output'), pwd / 'vasp_output')
                shutil.copy(poscar, pwd / 'CONTCAR')
            else:
                vasp_mock_output.append(
                    'MOCK PREPEND: Caller demanded to only locate test data by input, but no match was found.\n'
                )
                stop_and_return(vasp_mock_output)
    else:
        vasp_mock_output.append('MOCK PREPEND: Using test data from folder: ' + test_case + '\n')
        test_data_path = data_path(test_case, 'out')
        for out_object in pathlib.Path(test_data_path).iterdir():
            shutil.copy(out_object, pwd)

    # Read original vasp_output as we will append mock messages to it
    vasp_output_content = []
    if vasp_output_file.exists():
        with open(vasp_output_file, 'r', encoding='utf8') as handler:
            vasp_output_content = handler.readlines()

    vasp_mock_output.append('MOCK PREPEND: Mock folder contains the following files: ' + str(os.listdir(pwd)) + '\n')
    vasp_mock_output.append('MOCK PREPEND: END ----------------------\n')
    vasp_mock_output.append('Existing VASP stdout/stderr follows:\n')

    # Make sure we add the mock details in case we need to inspect later
    with open(vasp_output_file, 'w', encoding='utf8') as handler:
        handler.write(''.join(vasp_mock_output + vasp_output_content))


def stop_and_return(vasp_mock_output: List[str]) -> None:
    """Halts mock-vasp, rebuilds the vasp_output and returns."""
    # Assemble the
    print(''.join(vasp_mock_output))
    raise RuntimeError('The mock-vasp code could not perform a clean run.')



"""
Delegate types.

Module containing decorators and classes implementing delegate types.
"""

from __future__ import annotations

from functools import update_wrapper
from typing import Any, Callable, TypeVar

F = TypeVar('F', bound=Callable[..., Any])


def delegate_method_kwargs(prefix: str = '_init_with_') -> Callable[[F], F]:
    """
    Get a kwargs delegating decorator.

    :params prefix: (str) common prefix of delegate functions
    """

    def decorator(meth: F) -> F:
        """Decorate a class method to delegate kwargs."""

        def wrapper(*args: Any, **kwargs: Any) -> Any:
            for kwarg, value in kwargs.items():
                getattr(args[0], prefix + kwarg)(value)
            meth(*args, **kwargs)

        update_wrapper(wrapper, meth)
        return wrapper

    return decorator



"""
Representation of the POTCAR files.

Attempt to create a convenient but license-respecting storage system that also guarantees provenance.

Consists of two classes, PotcarData and PotcarFileData. Between the two data node classes exists a
one to one mapping but never a DbLink of any kind. The mapping must be defined in terms of a POTCAR
file hash sum.

Reasons for not using a file system based solution in general:

    * simplicity -> no necessity to define an fs based storage / retrieval schema
    * storage schema can be updated without manual user interaction
    * with fs based it is possible to lose enhanced provenance locally by deleting a file
    * This is easier to share between machines for same user / group members

Reasons for not using fs paths:

    * migrating to a new machine involves reinstating file hierarchy, might be non-trivial
    * corner cases with links, recursion etc

Reasons for not using pymatgen system:

    * changing an environment variable would invalidate provenance / disable reusing potentials
    * would block upgrading to newer pymatgen versions if they decide to change


Note::

    An fs based solution can be advantageous but should be 'expert mode' and not
    default solution due to provenance tradeoffs.

The following requirements have to be met:

    * The file hash attribute of PotcarFileData is unique in the Db
    * The file hash attribute of PotcarData is unique in the Db
    * Both classes can easily and quickly be found via the hash attribute
    * A PotcarData node can be exported without exporting the PotcarFileData node
    * The corresponding PotcarData node can be created at any time from the PotcarFileData node
    * PotcarFileData nodes are expected to be grouped in DbGroups called 'families'
    * The PotcarFileData nodes can be found according to their 'functional type' (pymatgen term)

The following would be nice to also allow optionally:

    * To pre-upload the files to a remote computer from a db and concat them right on there (to save traffic)
    * To use files directly on the remote computer (disclaimer: will never be as secure / tested)
    * To use existing pymatgen-style potentials library (disclaimer: support might break)

It is not to be expected for hundreds of distinct Potcar families to be present in the same database.

The mechanism for reading a POTCAR file into the Db::

    +-----------------------+
    [ parsing a POTCAR file ]
    +-----------------------+
            |
            v
            pmg_potcar = PotcarData.get_or_create_from_file()
            |
            v
     _----------------------------------------------_
    ( exists for PotcarFileData with pmg_potcar.sha512? )-----> no
     ^----------------------------------------------^         |
            |                                                 v
            v                                                 create
            yes                                               |
            |                                                 |
            v                                                 v
     _-------------------------_                             _-------------------------_
    ( Family given to parse to? ) -------------> no -+      ( Family given to parse to? )
     ^-------------------------^                     |       ^-------------------------^
            |                                        |        |         |
            v                                        |        |         no
            yes<------------------------------------]|[-------+         |
            |                                        |                  choose family according to functional type
            |                                        |                  |  (with fallback?)
            v                                        |                  |
            add existing PotcarFileData to family<--]|[-----------------+
            |                                        |
            |                     +------------------+
            v                     v
     _--------------------------------_
    ( exists corresponding PotcarData? )-----> no -----> create
     ^--------------------------------^ <------------------+
            |
            v
            return corresponding PotcarData

The mechanism for writing one or more PotcarData to file (from a calculation)::

    +-----------------------+
    [ Writing a POTCAR file ]
    +-----------------------+
            |
            v
            for each PotcarData node:
                get corresponding PotcarFileData <-> query for same symbol, family, hash, do not use links
            |
            v
            write_file using the write() of MultiPotcarIo

"""

# pylint: disable=import-outside-toplevel, too-many-lines
from __future__ import annotations, print_function

import hashlib
import os
import re
import shutil
import tarfile
import tempfile
from collections import namedtuple
from contextlib import contextmanager
from copy import deepcopy
from functools import cmp_to_key
from pathlib import Path
from typing import Any

from aiida.common import AIIDA_LOGGER
from aiida.common.exceptions import NotExistent, UniquenessError
from aiida.orm import (
    Data,  # pylint: disable=no-name-in-module
    Group,
    QueryBuilder,
)

from aiida_vasp.data.archive import ArchiveData
from aiida_vasp.utils.aiida_utils import get_current_user, querybuild
from aiida_vasp.utils.delegates import delegate_method_kwargs

# Records for the group type strings old and new
POTCAR_FAMILY_TYPE = 'vasp.potcar'
OLD_POTCAR_FAMILY_TYPE = 'data.vasp.potcar.family'


class PotcarGroup(Group):
    """
    A group for holding PotcarData objects that maps to various collections of POTCARs.
    """


def migrate_potcar_group() -> None:
    """
    Migrate existing potcar family groups to new specification.
    This creates copies of the old potcar family groups using the new `PotcarGroup` class.

    Despite the name 'migrate', the potcar family group created are in fact left as they are.
    """
    qdb = QueryBuilder()
    qdb.append(Group, filters={'type_string': OLD_POTCAR_FAMILY_TYPE})

    migrated = []
    created = []
    for (old_group,) in qdb.all():
        new_group, created = PotcarGroup.collection.get_or_create(
            label=old_group.label, description=old_group.description
        )
        new_group.add_nodes(list(old_group.nodes))
        new_group.store()
        migrated.append(new_group.label)
        if created:
            print(f'Created new style Group <{new_group}> for <{old_group.label}>')
        else:
            print(f'Adding nodes to existing new style Group <{new_group}> from <{old_group.label}>')


def normalize_potcar_contents(potcar_contents: str | bytes) -> str:
    """Normalize whitespace in a POTCAR given as a string."""
    try:
        potcar_contents = potcar_contents.decode()
    except AttributeError:
        pass
    normalized = re.sub(r'[ \t]+', r' ', potcar_contents)  # multiple spaces
    normalized = re.sub(r'[\n\r]\s*', r'\n', normalized)  # line breaks and spaces afterwards / empty lines
    normalized = re.sub(r'^\s*', r'', normalized)  # spaces / empty lines at the very beginning
    normalized = re.sub(r'\s*$', r'\n', normalized)  # trailing endline
    return normalized


def sha512_potcar(potcar_contents: str) -> str:
    """Hash the contents of a POTCAR file (given as str)."""
    sha512_hash = hashlib.sha512()
    sha512_hash.update(normalize_potcar_contents(potcar_contents).encode('utf-8'))
    return sha512_hash.hexdigest()


@contextmanager
def temp_dir() -> Any:
    """Temporary directory context manager that deletes the tempdir after use."""
    try:
        tempdir = tempfile.mkdtemp()
        yield Path(tempdir)
    finally:
        shutil.rmtree(tempdir)


@contextmanager
def temp_potcar(contents: bytes) -> Any:
    """Temporary POTCAR file from contents."""
    with temp_dir() as tempdir:
        potcar_file = tempdir / 'POTCAR'
        with potcar_file.open('wb') as potcar_fo:
            potcar_fo.write(contents)
        yield potcar_file


def extract_tarfile(file_path: Path) -> Path:
    """Extract a .tar archive into an appropriately named folder, return the path of the folder, avoid extracting if
    folder exists."""
    with tarfile.open(str(file_path)) as archive:
        new_dir = file_path.name.split('.tar')[0]
        new_path = file_path.parent / new_dir
        if not new_path.exists():
            archive.extractall(str(new_path))

    return new_path


def by_older(left: Any, right: Any) -> int:
    if left.ctime < right.ctime:
        return -1
    if left.ctime > right.ctime:
        return 1
    return 0


def by_user(left: Any, right: Any) -> int:
    if left.user.is_active and not right.user.is_active:
        return -1
    if not left.user.is_active and right.user.is_active:
        return 1
    return 0


class PotcarWalker:  # pylint: disable=useless-object-inheritance
    """
    Walk the file system and find POTCAR files under a given directory.

    Build a list of POTCARs including their full path and whether they are archived
    inside a tar archive.
    """

    def __init__(self, path: Path | str) -> None:  # pylint: disable=missing-function-docstring
        # Only accept a Path object or a string
        if isinstance(path, Path):
            self.path = path
        elif isinstance(path, str):
            self.path = Path(path)
        else:
            raise ValueError('The supplied path is not a Path object or a string.')
        self.potcars = set()

    def walk(self) -> None:
        """Walk the folder tree to find POTCAR, extracting any tar archives along the way."""
        if self.path.is_file():
            extracted = self.file_dispatch(self.path.parent, [], self.path.name)
            if extracted:
                self.path = extracted
                self.walk()
        else:
            for root, dirs, files in os.walk(str(self.path)):
                for file_name in files:
                    self.file_dispatch(root, dirs, file_name)

    def file_dispatch(self, root: str, dirs: list[str], file_name: str) -> Path | None:
        """Add POTCAR files to the list and dispatch handling of different kinds of files to other methods."""
        file_path = Path(root) / file_name
        if tarfile.is_tarfile(str(file_path)):
            return self.handle_tarfile(dirs, file_path)
        if 'POTCAR' in file_name:
            self.potcars.add(file_path)
        return None

    @classmethod
    def handle_tarfile(cls, dirs: list[str], file_path: Path) -> Path:
        """Handle .tar archives: extract and add the extracted folder to be searched."""
        new_dir = extract_tarfile(file_path)
        if str(new_dir) not in dirs:
            dirs.append(str(new_dir))
        return new_dir


class PotcarMetadataMixin:  # pylint: disable=useless-object-inheritance
    """Provide common Potcar metadata access and querying functionality."""

    _query_label = 'label'

    @classmethod
    def query_by_attrs(cls, query: Any = None, **kwargs: Any) -> Any:
        """Find a Data node by attributes."""
        label = cls._query_label
        if not query:
            query = querybuild(cls, tag=label)
        filters = {}
        for attr_name, attr_val in kwargs.items():
            filters[f'attributes.{attr_name}'] = {'==': attr_val}
        if cls._HAS_MODEL_VERSIONING:
            filters['attributes._MODEL_VERSION'] = {'==': kwargs.get('model_version', cls._VERSION)}
        query.add_filter(label, filters)
        return query

    @classmethod
    def find(cls, **kwargs: Any) -> list[Any]:
        """Find nodes by POTCAR metadata attributes given in kwargs."""
        query_builder = cls.query_by_attrs(**kwargs)
        if not query_builder.count():
            raise NotExistent()
        results = [result[0] for result in query_builder.all()]

        results.sort(key=cmp_to_key(by_older))
        return results

    @classmethod
    def find_one(cls, **kwargs: Any) -> Any:
        """
        Find one single node.

        Raise an exception if there are multiple.
        """
        res = cls.find(**kwargs)
        if len(res) > 1:
            if not all([True for node in res if node.sha512 == res[0].sha512]):  # pylint: disable=use-a-generator
                raise UniquenessError(f'Multiple nodes found satisfying {kwargs}')
        return res[0]

    @classmethod
    def exists(cls, **kwargs: Any) -> bool:
        """Answers the question wether a node with attributes given in kwargs exists."""
        return bool(cls.query_by_attrs(**kwargs).count() >= 1)

    @property
    def sha512(self) -> str:
        """Sha512 hash of the POTCAR file (readonly)."""
        return self.base.attributes.get('sha512')

    @property
    def title(self) -> str:
        """Title of the POTCAR file (readonly)."""
        return self.base.attributes.get('title')

    @property
    def functional(self) -> str:
        """Functional class of the POTCAR potential (readonly)."""
        return self.base.attributes.get('functional')

    @property
    def element(self) -> str:
        """Chemical element described by the POTCAR (readonly)."""
        return self.base.attributes.get('element')

    @property
    def symbol(self) -> str:
        """Element symbol property (VASP term) of the POTCAR potential (readonly)."""
        return self.base.attributes.get('symbol')

    @property
    def original_file_name(self) -> str:
        """The name of the original file uploaded into AiiDA."""
        return self.base.attributes.get('original_filename')

    @property
    def full_name(self) -> str:
        """The name of the original file uploaded into AiiDA."""
        return self.base.attributes.get('full_name')

    @property
    def potential_set(self) -> str:
        """The name of the original file uploaded into AiiDA."""
        return self.base.attributes.get('potential_set')

    def verify_unique(self) -> None:
        """Raise a UniquenessError if an equivalent node exists."""

        if self.exists(sha512=self.sha512):
            raise UniquenessError(f'A {self.__class__!s} node already exists for this file.')

        other_attrs = deepcopy(self.base.attributes.all)

        other_attrs.pop('sha512')
        if self.exists(**other_attrs):
            raise UniquenessError(
                f'A {self.__class__!s} node with these attributes but a different file exists:\n{other_attrs!s}'
            )


class VersioningMixin:  # pylint: disable=useless-object-inheritance
    """Minimalistic Node versioning."""

    _HAS_MODEL_VERSIONING = True
    _VERSION = None

    def set_version(self) -> None:
        self.base.attributes.set('_MODEL_VERSION', self._VERSION)

    @property
    def model_version(self) -> Any:
        return self.base.attributes.get('_MODEL_VERSION')

    @classmethod
    def old_versions_in_db(cls) -> bool:
        """Determine whether there are Nodes created with an older version of the model."""
        label = 'versioned'
        query = querybuild(cls, tag=label)
        filters = {'attributes._MODEL_VERSION': {'<': cls._VERSION}}
        query.add_filter(label, filters)
        return bool(query.count() >= 1)


class PotcarFileData(ArchiveData, PotcarMetadataMixin, VersioningMixin):
    """
    Store a POTCAR file in the db, never use as input to a calculation or workchain.

    .. warning:: Warning! Sharing nodes of this type may be illegal!

    In general POTCAR files may lay under license agreements, such as the ones distributed
    by the VASP group to VASP license holders. Take care to not share such licensed data
    with non-license holders.

    When writing a calculation plugin or workchain, do not use this as an input type,
    use :class:`aiida_vasp.data.potcar.PotcarData` instead!
    """

    _query_label = 'potcar_file'
    _query_type_string = 'data.vasp.potcar_file.'
    _plugin_type_string = 'data.vasp.potcar_file.PotcarFileData.'
    _VERSION = 1

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        # remove file in kwargs as this is not accepted in the subsequent inits
        path = kwargs.pop('file', None)
        super().__init__(*args, **kwargs)
        if path is not None:
            # Only allow a Path object or a string
            if isinstance(path, Path):
                self.init_with_kwargs(file=path)
            elif isinstance(path, str):
                self.init_with_kwargs(file=Path(path))
            else:
                raise ValueError('The supplied argument for file is not a Path object or a string.')

    @delegate_method_kwargs(prefix='_init_with_')
    def init_with_kwargs(self, **kwargs: Any) -> None:
        """Delegate initialization to _init_with - methods."""

    def _init_with_file(self, filepath: Path) -> None:
        """Initialized from a file path."""
        self.add_file(filepath)

    def add_file(self, src_abs: Path, dst_filename: Any = None) -> None:
        """Add the POTCAR file to the archive and set attributes."""
        from aiida_vasp.parsers.content_parsers.potcar import PotcarParser  # noqa: PLC0415

        self.set_version()
        if self._filelist:
            raise AttributeError('Can only hold one POTCAR file')
        super().add_file(src_abs, dst_filename)
        self.base.attributes.set('sha512', self.get_file_sha512(src_abs))
        with src_abs.open('r', encoding='utf8') as handler:
            potcar = PotcarParser(handler=handler)
        metadata = potcar.metadata
        self.base.attributes.set('title', metadata.titel)
        self.base.attributes.set('functional', metadata.functional)
        self.base.attributes.set('element', metadata.element)
        self.base.attributes.set('symbol', metadata.symbol)
        src_path = src_abs.resolve()
        src_rel = src_path.relative_to(src_path.parents[2])  # familyfolder/Element/POTCAR
        # Make sure we store string elements of Path in the attributes
        self.base.attributes.set('original_filename', str(src_rel))
        dir_name = src_path.parent
        dir_name = dir_name.name
        self.base.attributes.set('full_name', str(dir_name))
        self.base.attributes.set('potential_set', str(src_path.parts[-3]))

    @classmethod
    def get_file_sha512(cls, path: Path | str) -> str:
        """Get the sha512 sum for a POTCAR file (after whitespace normalization)."""
        path = Path(path)
        with path.open('r', encoding='utf8') as potcar_fo:
            sha512 = sha512_potcar(potcar_fo.read())
        return sha512

    @classmethod
    def get_contents_sha512(cls, contents: str) -> str:
        """Get the sha512 sum for the contents of a POTCAR file (after normalization)."""
        return sha512_potcar(contents)

    # pylint: disable=arguments-differ
    def store(self, *args: Any, **kwargs: Any) -> Any:
        """Ensure uniqueness and existence of a matching PotcarData node before storing."""
        self.set_version()
        _ = PotcarData.get_or_create(self)
        self.verify_unique()
        return super().store(*args, **kwargs)

    @contextmanager
    def get_file_obj(self) -> Any:
        """Open a readonly file object to read the stored POTCAR file."""
        file_obj = None
        with self.get_archive() as archive:
            try:
                file_obj = archive.extractfile(archive.members[0])
                yield file_obj
            finally:
                if file_obj:
                    file_obj.close()

    @contextmanager
    def get_file_obj_and_tar_obj(self) -> Any:
        """Return both decompressed file object and the archive object"""
        file_obj = None
        with self.get_archive() as archive:
            try:
                file_obj = archive.extractfile(archive.members[0])
                yield file_obj, archive
            finally:
                if file_obj:
                    file_obj.close()

    def export_archive(self, archive: Any, dry_run: bool = False) -> str:
        """Add the stored POTCAR file to an archive for export."""
        with self.get_file_obj_and_tar_obj() as objects:
            potcar_fo, tar_fo = objects
            arcname = f'{self.symbol}/POTCAR'
            tarinfo = tar_fo.members[0]
            tarinfo.name = arcname
            if not dry_run:
                archive.addfile(tarinfo, fileobj=potcar_fo)
        return tarinfo.name

    def export_file(self, path: Path, dry_run: bool = False) -> Path:
        """
        Write the contents of the stored POTCAR file to a destination on the local file system.

        :param path: path to the destination file or folder as a Path or string object

        When given a folder, the destination file will be created in a subdirectory with the name of the symbol.
        This is for conveniently exporting multiple files into the same folder structure as the POTCARs are
        distributed in.

        Examples::

            potcar_file = PotcarFileData.get_or_create(<file>)
            assert potcar_file.symbol == 'Si_d'

            potcar_file.export('./POTCAR.Si')
            ## writes to ./POTCAR.Si

            potcar_file.export('./potcars/')
            ## writes to
            ## ./
            ##  |-potcars/
            ##           |-Si_d/
            ##                 |-POTCAR
        """
        path = Path(path)
        if path.is_dir():
            path = path / self.symbol / 'POTCAR'
        if not dry_run:
            # Make sure the directory exists
            path_dir = path.parent
            path_dir.mkdir(parents=True, exist_ok=True)
            with path.open(mode='wb') as dest_fo:
                dest_fo.write(self.get_content())
        return path

    def get_content(self) -> bytes:
        with self.get_file_obj() as potcar_fo:
            return potcar_fo.read()

    @classmethod
    def get_or_create(cls, filepath: Path) -> tuple[Any, bool]:
        """Get or create (store) a PotcarFileData node."""
        sha512 = cls.get_file_sha512(filepath)
        if cls.exists(sha512=sha512):
            created = False
            node = cls.find_one(sha512=sha512)
        else:
            created = True
            node = cls(file=filepath)
            node.store()
        return node, created

    @classmethod
    def get_or_create_from_contents(cls, contents: str) -> tuple[Any, bool]:
        """Get or create (store) a PotcarFileData node from a string containing the POTCAR contents."""
        with temp_potcar(contents) as potcar_file:
            return cls.get_or_create(potcar_file)


class PotcarData(Data, PotcarMetadataMixin, VersioningMixin):
    """
    Store enough metadata about a POTCAR file to identify and find it.

    Meant to be used as an input to calculations. This node type holds no
    licenced data and can be freely shared without legal repercussions.
    """

    _query_label = 'potcar'
    _query_type_string = 'data.vasp.potcar.'
    _plugin_type_string = 'data.vasp.potcar.PotcarData.'
    _VERSION = 1

    def __init__(self, **kwargs: Any) -> None:
        potcar_file_node = kwargs.pop('potcar_file_node', None)
        super().__init__(**kwargs)
        if potcar_file_node is not None:
            self.set_potcar_file_node(potcar_file_node)

    def set_potcar_file_node(self, potcar_file_node: Any) -> None:
        """Initialize from a PotcarFileData node."""
        self.set_version()
        for attr_name in potcar_file_node.base.attributes.all.keys():
            self.base.attributes.set(attr_name, potcar_file_node.base.attributes.get(attr_name))

    def find_file_node(self) -> Any:
        """Find and return the matching PotcarFileData node."""
        return PotcarFileData.find_one(sha512=self.sha512)

    # pylint: disable=arguments-differ,signature-differs
    def store(self, *args: Any, **kwargs: Any) -> Any:
        """Ensure uniqueness before storing."""
        self.set_version()
        self.verify_unique()
        return super().store(*args, **kwargs)

    @classmethod
    def get_or_create(cls, file_node: Any) -> tuple[Any, bool]:
        """Get or create (store) a PotcarData node."""

        created = False
        try:
            node = cls.find_one(sha512=file_node.sha512)
        except NotExistent:
            node = cls(potcar_file_node=file_node)
            created = True
            node.store()
        return node, created

    @classmethod
    def get_or_create_from_file(cls, file_path: str | Path) -> tuple[Data, bool]:
        """Get or create (store) a PotcarData node from a POTCAR file."""
        sha512 = PotcarFileData.get_file_sha512(file_path)
        try:
            file_node = PotcarFileData.find_one(sha512=sha512)
        except NotExistent:
            file_node = PotcarFileData(file=file_path)
        node, created = cls.get_or_create(file_node)
        if not file_node.is_stored:
            file_node.store()
        return node, created

    @classmethod
    def get_or_create_from_contents(cls, contents: str) -> tuple[Data, bool]:
        """Get or create (store) a PotcarData node from a string containing the POTCAR contents."""
        with temp_potcar(contents) as potcar_file:
            return cls.get_or_create_from_file(str(potcar_file))

    @classmethod
    def file_not_uploaded(cls, file_path: str | Path) -> PotcarFileData | tuple:
        sha512 = PotcarFileData.get_file_sha512(file_path)
        return (
            PotcarFileData.find_one(sha512=sha512)
            if PotcarFileData.exists(sha512=sha512)
            else namedtuple('potcar', ('uuid'))('-1')
        )

    def get_family_names(self) -> list[str]:
        """List potcar families to which this instance belongs."""
        return [group.label for group in PotcarGroup.query(nodes=self)]

    @classmethod
    def get_potcar_group(cls, group_name: str) -> PotcarGroup | None:
        """Return the PotcarFamily group with the given name."""
        try:
            group = PotcarGroup.collection.get(label=group_name)
        except NotExistent:
            group = None
        return group

    @classmethod
    def get_potcar_groups(
        cls, filter_elements: list[str] | str | None = None, filter_symbols: list[str] | None = None
    ) -> list[PotcarGroup]:
        """
        List all names of groups of type PotcarFamily, possibly with some filters.

        :param filter_elements: list of strings.
               If present, returns only the groups that contains one POTCAR for
               every element present in the list. Default=None, meaning that
               all families are returned. A single element can be passed as a string.
        :param filter_symbols: list of strings with symbols to filter for.
        """
        group_query = QueryBuilder()
        group_query.append(PotcarGroup, with_node='potcar_data', tag='potcar_data', project='*')

        groups = [group_list[0] for group_list in group_query.all()]

        if filter_elements:
            for element in filter_elements:
                idx_has_element = []
                for i, group in enumerate(groups):
                    elem_query = QueryBuilder()
                    elem_query.append(PotcarGroup, tag='family', filters={'label': {'==': group.label}})
                    elem_query.append(
                        cls, tag='potcar', with_group='family', filters={'attributes.element': {'==': element}}
                    )
                    if elem_query.count() > 0:
                        idx_has_element.append(i)
                groups = [groups[i] for i in range(len(groups)) if i in idx_has_element]

        if filter_symbols:
            for symbol in filter_symbols:
                idx_has_symbol = []
                for i, group in enumerate(groups):
                    symbol_query = QueryBuilder()
                    symbol_query.append(PotcarGroup, tag='family', filters={'label': {'==': group.label}})
                    symbol_query.append(
                        cls, tag='potcar', with_group='family', filters={'attributes.symbol': {'==': symbol}}
                    )
                    if symbol_query.count() > 0:
                        idx_has_symbol.append(i)
                groups = [groups[i] for i in range(len(groups)) if i in idx_has_symbol]

        return groups

    @classmethod
    def get_potcars_dict(
        cls, elements: list[str], family_name: str, mapping: dict[str, str] | None = None, auto_migrate: bool = True
    ) -> dict[str, Any]:
        """
        Get a dictionary {element: ``PotcarData.full_name``} for all given symbols.

        :param elements: The list of symbols to find POTCARs for
        :param family_name: The POTCAR family to be used
        :param mapping: A mapping[element] -> ``full_name``, for example: mapping={'In': 'In', 'As': 'As_d'}
        :param auto_migrate: A flag of whether to perform the migration automatically when
          migration is found to be needed.

        Exceptions:

         *If the mapping does not contain an item for a given element name, raise a ``ValueError``.
         *If no POTCAR is found for a given element, a ``NotExistent`` error is raised.

        If there are multiple POTCAR with the same ``full_name``, the first one
        returned by ``PotcarData.find()`` will be used.
        """
        if not mapping:
            mapping = {element: element for element in elements}
        group_filters = {'label': {'==': family_name}}
        element_filters = {'attributes.full_name': {'in': [mapping[element] for element in elements]}}
        query = QueryBuilder()
        query.append(PotcarGroup, tag='family', filters=group_filters)
        query.append(cls, tag='potcar', with_group='family', filters=element_filters)

        result_potcars = {}
        for element in elements:
            if element not in mapping:
                raise ValueError(
                    'Potcar mapping must contain an item for each element in the structure, '
                    'with the full name of the POTCAR file (i.e. "In_d", "As_h").'
                )
            full_name = mapping[element]
            potcars_of_kind = [potcar[0] for potcar in query.all() if potcar[0].full_name == full_name]
            if not potcars_of_kind:
                # Check if it was because the family has not been migrated
                query = QueryBuilder()
                query.append(Group, filters={'type_string': OLD_POTCAR_FAMILY_TYPE, 'label': family_name}, tag='family')
                query.append(cls, tag='potcar', with_group='family', filters=element_filters)
                if query.count() > 1:
                    if auto_migrate:
                        # Migrate to new group labels, and retry
                        migrate_potcar_group()
                        return cls.get_potcars_dict(
                            elements=elements, family_name=family_name, mapping=mapping, auto_migrate=False
                        )
                    raise NotExistent(
                        (
                            'No POTCAR found for full name {} in family {}, but it was found in a legacy '
                            'group with the same name.'
                            ' Please run `verdi data vasp.potcar migratefamilies`.'
                        ).format(full_name, family_name)
                    )

                raise NotExistent(f'No POTCAR found for full name {full_name} in family {family_name}')
            if len(potcars_of_kind) > 1:
                result_potcars[element] = cls.find(family=family_name, full_name=full_name)[0]
            else:
                result_potcars[element] = potcars_of_kind[0]

        return result_potcars

    @classmethod
    def query_by_attrs(cls, query: Any = None, **kwargs: Any) -> Any:
        family_name = kwargs.pop('family_name', None)
        if family_name:
            group_filters = {'label': {'==': family_name}}
            query = QueryBuilder()
            query.append(PotcarGroup, tag='family', filters=group_filters)
            query.append(cls, tag=cls._query_label, with_group='family')
        return super(PotcarData, cls).query_by_attrs(query=query, **kwargs)

    @classmethod
    def get_full_names(cls, family_name: str | None = None, element: str | None = None) -> list[str]:
        """
        Gives a set of symbols provided by this family.

        Not every symbol may be supported for every element.
        """
        query = cls.query_by_attrs(family_name=family_name, element=element)
        query.add_projection(cls._query_label, 'attributes.full_name')
        return [name[0] for name in query.all()]

    @classmethod
    def get_potcars_from_structure(
        cls, structure: Any, family_name: str, mapping: dict[str, str] | None = None
    ) -> dict[str, Any]:
        """
        Given a POTCAR family name and a AiiDA structure, return a dictionary associating each kind name with
        its PotcarData object.

        :param structure: An AiiDA structure
        :param family_name: The POTCAR family to be used
        :param mapping: A mapping[kind name] -> ``full_name``, for example: mapping={'In1': 'In', 'In2': 'In_d'}

        The Dictionary looks as follows::

            {
                kind1.name: PotcarData_for_kind1,
                kind2.name: ...
            }

        This is to make the output of this function suitable for giving directly as input
        to VaspCalculation.process() instances.

        :raise MultipleObjectsError: if more than one UPF for the same element is found in the group.
        :raise NotExistent: if no UPF for an element in the group is found in the group.


        Example::

            ## using VASP recommended POTCARs
            from aiida_vasp.utils.default_paws import DEFAULT_LDA, DEFAULT_GW
            vasp_process = CalculationFactory('vasp.vasp').process()
            inputs = vasp_process.get_inputs_template()
            inputs.structure = load_node(123)
            inputs.potential = PotcarData.get_potcars_from_structure(
                structure=inputs.structure,
                family_name='PBE',
                mapping=DEFAULT_GW
            )

            ## using custom POTCAR map
            custom_mapping = {
                'In1': 'In',
                'In2': 'In_d',
                'As': 'As_d'
            }
            inputs.potential = PotcarData.get_potcars_from_structure(
                structure=inputs.structure,
                family_name='PBE',
                mapping=custom_mapping
            )
        """
        kind_names = structure.get_kind_names()
        potcar_dict = dict(cls.get_potcars_dict(kind_names, family_name, mapping=mapping))
        return potcar_dict

    @classmethod
    def _prepare_group_for_upload(
        cls, group_name: str, group_description: str | None = None, dry_run: bool = False
    ) -> PotcarGroup:
        """Prepare a (possibly new) group to upload a POTCAR family to."""
        if not dry_run:
            group, group_created = PotcarGroup.collection.get_or_create(label=group_name)
        else:
            group = cls.get_potcar_group(group_name)
            group_created = bool(not group)
            if not group:
                group = Group(label=group_name)

        if group.user.pk != get_current_user().pk:
            raise UniquenessError(
                f'There is already a POTCAR family group with name {group_name}, '
                f'but it belongs to user {group.user.email},'
                ' therefore you cannot modify it.'
            )

        if group_description:
            group.description = group_description
        elif group_created:
            raise ValueError(f'A new POTCAR family {group_name} should be created but no description was given!')

        return group

    @classmethod
    def upload_potcar_family(
        cls,
        source: str | Path,
        group_name: str,
        group_description: str | None = None,
        stop_if_existing: bool = True,
        dry_run: bool = False,
    ) -> tuple[int, int, int]:
        """
        Upload a set of POTCAR potentials as a family.

        :param source: a path containing all POTCAR files to be added.
        :param group_name: the name of the group to create. If it exists and is
            non-empty, a UniquenessError is raised.
        :param group_description: a string to be set as the group description.
            Overwrites previous descriptions, if the group was existing.
        :param stop_if_existing: if True, check for the sha512 of the files and,
            if the file already exists in the DB, raises a MultipleObjectsError.
            If False, simply adds the existing UPFData node to the group.
        :param dry_run: If True, do not change the database.
        """
        group = cls._prepare_group_for_upload(group_name, group_description, dry_run=dry_run)

        potcar_finder = PotcarWalker(source)
        potcar_finder.walk()
        num_files = len(potcar_finder.potcars)
        family_nodes_uuid = [node.uuid for node in group.nodes] if not dry_run else []
        potcars_tried_upload = cls._try_upload_potcars(
            potcar_finder.potcars, stop_if_existing=stop_if_existing, dry_run=dry_run
        )
        new_potcars_added = [
            (potcar, created, file_path)
            for potcar, created, file_path in potcars_tried_upload
            if potcar.uuid not in family_nodes_uuid
        ]

        for potcar, created, file_path in new_potcars_added:
            if created:
                AIIDA_LOGGER.debug(
                    'New PotcarData node %s created while uploading file %s for family %s',
                    potcar.uuid,
                    file_path,
                    group_name,
                )
            else:
                AIIDA_LOGGER.debug(
                    'PotcarData node %s used instead of uploading file %s to family %s',
                    potcar.uuid,
                    file_path,
                    group_name,
                )

        if not dry_run:
            group.add_nodes([potcar for potcar, created, file_path in new_potcars_added])

        num_added = len(new_potcars_added)
        num_uploaded = len([item for item in new_potcars_added if item[1]])  # item[1] refers to 'created'

        return num_files, num_added, num_uploaded

    @classmethod
    def _try_upload_potcars(
        cls, file_paths: list[Path], stop_if_existing: bool = True, dry_run: bool = False
    ) -> list[tuple[Any, bool, str]]:
        """Given a list of absolute paths to potcar files, try to upload them (or pretend to if dry_run=True)."""
        list_created = []
        for file_path_obj in file_paths:
            file_path = str(file_path_obj)
            try:
                if not dry_run:
                    potcar, created = cls.get_or_create_from_file(file_path)
                else:
                    potcar = cls.file_not_uploaded(file_path)
                    created = bool(potcar.uuid == -1)
                if stop_if_existing and not created:
                    raise ValueError(
                        (
                            'A POTCAR with identical SHA512 to {} is already in the DB,'
                            'therefore it cannot be added with the stop_if_existing kwarg.'
                        ).format(file_path)
                    )
                list_created.append((potcar, created, file_path))
            except KeyError as err:
                print(f'skipping file {file_path} - uploading raised {err.__class__!s}{err!s}')
            except AttributeError as err:
                print(f'skipping file {file_path} - uploading raised {err.__class__!s}{err!s}')
            except IndexError as err:
                print(f'skipping file {file_path} - uploading raised {err.__class__!s}{err!s}')

        return list_created

    @classmethod
    def export_family_folder(
        cls, family_name: str, path: str | Path | None = None, dry_run: bool = False
    ) -> list[Path]:
        """
        Export a family of POTCAR nodes into a file hierarchy similar to the one POTCARs are distributed in.

        :param family_name: name of the POTCAR family
        :param path: path to a local directory, either a string or Path object, default to current directory
        :param dry_run: bool, if True, only collect the names of files that would otherwise be written.

        If ``path`` already exists, everything will be written into a subdirectory with the name of the family.
        """
        # Only allow Path or string
        if path is not None:
            if isinstance(path, (Path, str)):
                path = Path(path)
            else:
                raise ValueError('The supplied path is not a Path object or a string.')
        else:
            path = Path()

        if path.exists():
            path = path / family_name
        group = cls.get_potcar_group(family_name)
        all_file_nodes = [potcar.find_file_node() for potcar in group.nodes]
        files_written = []

        with temp_dir() as staging_dir:
            for file_node in all_file_nodes:
                new_file = file_node.export_file(staging_dir, dry_run=dry_run)
                files_written.append(path / new_file.relative_to(staging_dir))
            if not dry_run:
                # copytree uses copy2 which conserves all metadata as well
                shutil.copytree(staging_dir, path)

        return files_written

    @classmethod
    def export_family_archive(
        cls, family_name: str, path: str | Path | None = None, dry_run: bool = False
    ) -> tuple[Path, list[str]]:
        """Export a family of POTCAR nodes into a compressed archive."""
        # Only allow Path or string
        if path is not None:
            if isinstance(path, (Path, str)):
                path = Path(path)
            else:
                raise ValueError('The supplied path is not a Path object or a string.')
        else:
            path = Path()

        if path.is_dir():
            path = path / family_name

        if not path.suffix:
            name = path.name + '.tar.gz'
            path = path.parent / name

        group = cls.get_potcar_group(family_name)
        all_file_nodes = [potcar.find_file_node() for potcar in group.nodes]
        files_added = []
        if not dry_run:
            with tarfile.open(str(path), 'w:gz') as archive:
                for file_node in all_file_nodes:
                    files_added.append(file_node.export_archive(archive, dry_run=dry_run))
        else:
            for file_node in all_file_nodes:
                files_added.append(file_node.export_archive(None, dry_run=dry_run))

        return path, files_added

    def get_content(self) -> bytes:
        return self.find_file_node().get_content()

    @classmethod
    def find(cls, **kwargs: Any) -> list[Any]:
        """
        Extend :py:meth:`PotcarMetadataMixin.find` with filtering by POTCAR family.

        If no POTCAR is found, raise a ``NotExistent`` exception.

        If multiple POTCAR are found, sort them by:

            * POTCARS belonging to the active user first
            * oldest first
        """
        family = kwargs.pop('family', None)
        if not family:
            return super(PotcarData, cls).find(**kwargs)
        query = cls.query_by_attrs(**kwargs)
        group_filters = {'label': {'==': family}}
        query.append(PotcarGroup, tag='family', filters=group_filters, with_node=cls._query_label)
        query.add_projection(cls._query_label, '*')
        if not query.count():
            raise NotExistent()
        results = [result[0] for result in query.all()]

        results.sort(key=cmp_to_key(by_older))
        return results



"""
Base calculation class.

Base and meta classes for VASP calculation classes.
"""

from __future__ import annotations

# pylint: disable=abstract-method,invalid-metaclass,ungrouped-imports
# explanation: pylint wrongly complains about Node not implementing query
import os
from typing import TYPE_CHECKING, Any

from aiida import orm
from aiida.common import CalcInfo, CodeInfo, ValidationError
from aiida.engine import CalcJob

if TYPE_CHECKING:
    from aiida.common.folders import Folder
    from aiida.engine.processes.calcjobs.calcjob import CalcJobProcessSpec


class VaspCalcBase(CalcJob):
    """
    Base class of all calculations utilizing VASP.

    * Defines internal parameters common to all vasp calculations.
    * provides a basic, extendable implementation of _prepare_for_submission
    * provides hooks, so subclasses can extend the behavior without having to reimplement common functionality
    """

    _default_parser = 'vasp.vasp'

    @classmethod
    def define(cls, spec: CalcJobProcessSpec) -> None:
        super(VaspCalcBase, cls).define(spec)
        spec.input(
            'restart_folder',
            valid_type=orm.RemoteData,
            help='A remote folder to restart from if need be',
            required=False,
        )

    @classmethod
    def max_retrieve_list(cls) -> list[str]:
        """Return a list of all possible output objects from a VASP run."""
        retrieve_list = [
            'CHG',
            'CHGCAR',
            'AECCAR0',
            'AECCAR1',
            'AECCAR2',
            'ELFCAR',
            'PARCHG',  # Density related
            'CONTCAR',
            'XDATCAR',
            'PCDAT',  # Structure related
            'DOSCAR',
            'EIGENVAL',
            'PROCAR',  # Electronic structure related
            'IBZKPT',  # Irreducible k-points
            'LOCPOT',  # Potential related
            'BSEFATBAND',  # Eigenvectors of the BSE matrix
            'WAVECAR',
            'WAVEDER',
            'PROOUT',
            'TMPCAR',
            'W*.tmp',
            'WFULL*.tmp',  # Wave function related properties
            'wannier90*',  # Wannier90 related
            'OSZICAR',  # Convergence related
            'REPORT',  # Output of molecular dynamics runs
            'STOPCAR',  # Controlled stopping
            'vasprun.xml',
            'OUTCAR',
        ]
        return retrieve_list

    def prepare_for_submission(self, folder: Folder) -> CalcInfo:  # pylint: disable=arguments-differ
        """
        Writes the four minimum outputs: INCAR, POSCAR, POTCAR, KPOINTS.

        Delegates the construction and writing / copying to write_<object> methods.
        That way, subclasses can use any form of input nodes and just
        have to implement the write_xxx method accordingly.
        Subclasses can extend by calling the super method and if necessary
        modifying it's output CalcInfo before returning it.
        """
        # write input objects
        incar = folder.get_abs_path('INCAR')
        structure = folder.get_abs_path('POSCAR')
        potentials = folder.get_abs_path('POTCAR')
        kpoints = folder.get_abs_path('KPOINTS')

        remote_copy_list = []

        self.verify_inputs()
        if self._is_restart():
            remote_copy_list.extend(self.remote_copy_restart_folder())
        self.write_incar(incar)
        self.write_poscar(structure)
        self.write_potcar(potentials)
        self.write_kpoints(kpoints)

        # calcinfo
        calcinfo = CalcInfo()
        calcinfo.uuid = self.uuid
        calcinfo.retrieve_list = self.max_retrieve_list()
        codeinfo = CodeInfo()
        codeinfo.code_uuid = self.inputs.code.uuid
        codeinfo.code_pk = self.inputs.code.pk
        calcinfo.codes_info = [codeinfo]
        calcinfo.remote_copy_list = remote_copy_list
        # here we need to do the charge density and wave function copy
        # as we need access to the calcinfo
        calcinfo.local_copy_list = []
        self.write_additional(folder, calcinfo)

        return calcinfo

    def remote_copy_restart_folder(self) -> list[tuple[str, str, str]]:
        """Add all objects required for restart to the list of objects to be copied from the previous calculation."""
        restart_folder = self.inputs.restart_folder
        computer = self.node.computer
        included = ['CHGCAR', 'WAVECAR']
        included += self.inputs.settings.base.attributes.get('ADDITIONAL_REMOTE_COPY_LIST', [])
        existing_objects = restart_folder.listdir()
        to_copy = []
        for name in included:
            if name not in existing_objects:
                # Here we simple issue an warning as the requirement of objects will be explicitly checked by
                # `write_additional` method
                self.report(f'WARNING: Object {name} does not exist in the restart folder.')
            else:
                to_copy.append(name)
        copy_list = [(computer.uuid, os.path.join(restart_folder.get_remote_path(), name), '.') for name in to_copy]
        return copy_list

    def verify_inputs(self) -> bool:
        """
        Hook to be extended by subclasses with checks for input nodes.

        Is called once before submission.
        """
        self.check_restart_folder()
        return True

    def check_restart_folder(self) -> None:
        restart_folder = self.inputs.get('restart_folder', None)
        if restart_folder:
            if not self.node.computer.pk == restart_folder.computer.pk:
                raise ValidationError('Calculation can not be restarted on another computer')

    def _is_restart(self) -> bool:
        restart_folder = self.inputs.get('restart_folder', None)
        return bool(restart_folder)

    def store(self, *args: Any, **kwargs: Any) -> None:
        """Adds a _prestore subclass hook for operations that should be done just before storing."""
        self._prestore()
        super().store(*args, **kwargs)

    def _prestore(self) -> None:
        """Subclass hook for updating attributes etc, just before storing."""
        return

    def write_additional(self, folder: Folder, calcinfo: CalcInfo) -> None:  # pylint: disable=unused-argument,
        """Subclass hook to write additional input objects."""
        return



"""
The ``vasprun.xml`` parser interface.

Contains the parsing interfaces to ``parsevasp`` used to parse ``vasprun.xml`` content.
"""

# pylint: disable=abstract-method, too-many-public-methods
from typing import Any, BinaryIO, TextIO

import numpy as np
from parsevasp import constants as parsevaspct
from parsevasp.vasprun import Xml

from aiida_vasp.parsers.content_parsers.base import BaseFileParser
from aiida_vasp.utils.compare_bands import get_band_properties


class VasprunParser(BaseFileParser):
    """The parser interface that enables parsing of ``vasprun.xml`` content.

    The parser is triggered by using the keys listed in ``PARSABLE_QUANTITIES``.

    """

    OPEN_MODE = 'rb'
    DEFAULT_SETTINGS = {
        'quantities_to_parse': [
            'structure',
            'eigenvalues',
            'dos',
            'kpoints',
            'occupancies',
            'trajectory',
            'energies',
            'projectors',
            'dielectrics',
            'born_charges',
            'hessian',
            'dynmat',
            'forces',
            'stress',
            'total_energies',
            'band_properties',
            'version',
            'parameters',
        ],
        'energy_type': ['energy_extrapolated'],
        'electronic_step_energies': False,
    }

    PARSABLE_QUANTITIES = {
        'structure': {'inputs': [], 'name': 'structure', 'prerequisites': [], 'alternatives': ['poscar-structure']},
        'eigenvalues': {
            'inputs': [],
            'name': 'eigenvalues',
            'prerequisites': [],
            'alternatives': ['eigenval-eigenvalues'],
        },
        'dos': {'inputs': [], 'name': 'dos', 'prerequisites': [], 'alternatives': ['doscar-dos']},
        'kpoints': {'inputs': [], 'name': 'kpoints', 'prerequisites': [], 'alternatives': ['kpoints-kpoints']},
        'occupancies': {
            'inputs': [],
            'name': 'occupancies',
            'prerequisites': [],
        },
        'trajectory': {
            'inputs': [],
            'name': 'trajectory',
            'prerequisites': [],
        },
        'energies': {
            'inputs': [],
            'name': 'energies',
            'prerequisites': [],
        },
        'total_energies': {
            'inputs': [],
            'name': 'total_energies',
            'prerequisites': [],
        },
        'projectors': {
            'inputs': [],
            'name': 'projectors',
            'prerequisites': [],
        },
        'dielectrics': {
            'inputs': [],
            'name': 'dielectrics',
            'prerequisites': [],
        },
        'stress': {
            'inputs': [],
            'name': 'stress',
            'prerequisites': [],
        },
        'forces': {
            'inputs': [],
            'name': 'forces',
            'prerequisites': [],
        },
        'born_charges': {
            'inputs': [],
            'name': 'born_charges',
            'prerequisites': [],
        },
        'hessian': {
            'inputs': [],
            'name': 'hessian',
            'prerequisites': [],
        },
        'dynmat': {
            'inputs': [],
            'name': 'dynmat',
            'prerequisites': [],
        },
        'fermi_level': {
            'inputs': [],
            'name': 'fermi_level',
            'prerequisites': [],
        },
        'band_properties': {
            'inputs': [],
            'name': 'band_properties',
            'prerequisites': [],
        },
        'version': {
            'inputs': [],
            'name': 'version',
            'prerequisites': [],
        },
        'parameters': {
            'inputs': [],
            'name': 'parameters',
            'prerequisites': [],
        },
    }

    # Mapping of the energy names to those returned by parsevasp.vasprunl.Xml
    ENERGY_MAPPING = {
        'energy_extrapolated': 'energy_extrapolated_final',
        'energy_free': 'energy_free_final',
        'energy_no_entropy': 'energy_no_entropy_final',
        'energy_extrapolated_electronic': 'energy_extrapolated',
        'energy_free_electronic': 'energy_free',
        'energy_no_entropy_electronic': 'energy_no_entropy',
    }

    ENERGY_MAPPING_VASP5 = {
        'energy_extrapolated': 'energy_no_entropy_final',
        'energy_free': 'energy_free_final',
        # Not that energy_extrapolated_final parsed is the entropy term
        'energy_no_entropy': 'energy_extrapolated_final',
        'energy_extrapolated_electronic': 'energy_extrapolated',
        'energy_free_electronic': 'energy_free',
        'energy_no_entropy_electronic': 'energy_no_entropy',
    }

    def _init_from_handler(self, handler: BinaryIO | TextIO) -> None:
        """Initialize using a file like handler."""

        self.overflow = False
        try:
            self._content_parser = Xml(file_handler=handler, k_before_band=True, logger=self._logger)
        except SystemExit as exception:
            if exception.code == 509:
                # Xml might be fine but overflow is detected
                self.overflow = True
                self.parser_notifications['vasprun_xml_overflow'] = True
                self._logger.warning('Parsevasp exited abnormally due to overflow in XML file.')
            else:
                self._logger.warning('Parsevasp exited abnormally.')

    @property
    def version(self) -> str | None:
        """Fetch the VASP version from ``parsevasp`` and return it as a string object."""

        # fetch version
        version = self._content_parser.get_version()

        if version is None:
            return None

        return version

    @property
    def eigenvalues(self) -> dict[str, Any] | None:
        """Fetch eigenvalues."""

        # Fetch eigenvalues
        eigenvalues = self._content_parser.get_eigenvalues()

        if eigenvalues is None:
            return None

        return eigenvalues

    @property
    def occupancies(self) -> dict[str, Any] | None:
        """Fetch occupancies."""

        # Fetch occupancies
        occupancies = self._content_parser.get_occupancies()

        if occupancies is None:
            # occupancies not present, should not really happen?
            return None

        return occupancies

    @property
    def kpoints(self) -> dict[str, Any] | None:
        """Fetch the kpoints an prepare for consumption by the NodeComposer."""

        kpts = self._content_parser.get_kpoints()
        kptsw = self._content_parser.get_kpointsw()
        # k-points in XML is always in reciprocal if spacing methods have been used
        # but what about explicit/regular
        cartesian = False
        kpoints_data = None
        if (kpts is not None) and (kptsw is not None):
            # Create a dictionary and store k-points that can be consumed by the NodeComposer
            kpoints_data = {}
            kpoints_data['mode'] = 'explicit'
            kpoints_data['cartesian'] = cartesian
            kpoints_data['points'] = kpts
            kpoints_data['weights'] = kptsw

        return kpoints_data

    @property
    def structure(self) -> dict[str, Any] | None:
        """
        Fetch a given structure.

        Which structure to fetch is controlled by inputs.

        eFL: Need to clean this so that we can set different
        structures to pull from the outside. Could be usefull not
        pulling the whole trajectory.

        Currently defaults to the last structure.

        """

        return self.last_structure

    @property
    def last_structure(self) -> dict[str, Any] | None:
        """
        Fetch the structure.

        After or at the last recorded ionic step.

        """

        last_lattice = self._content_parser.get_lattice('last')
        if last_lattice is None:
            return None
        return _build_structure(last_lattice)

    @property
    def final_structure(self) -> dict[str, Any] | None:
        """
        Fetch the structure.

        After or at the last recorded ionic step. Should in
        principle be the same as the method above.

        """

        return self.last_structure

    @property
    def last_forces(self) -> np.ndarray | None:
        """
        Fetch forces.

        After or at the last recorded ionic step.

        """

        force = self._content_parser.get_forces('last')
        return force

    @property
    def final_forces(self) -> np.ndarray | None:
        """
        Fetch forces.

        After or at the last recorded ionic step.

        """

        return self.last_forces

    @property
    def forces(self) -> np.ndarray | None:
        """
        Fetch final forces.
        """
        return self.final_forces

    @property
    def last_stress(self) -> np.ndarray | None:
        """
        Fetch stess.

        After or at the last recorded ionic step.

        """

        stress = self._content_parser.get_stress('last')
        return stress

    @property
    def final_stress(self) -> np.ndarray | None:
        """
        Fetch stress.

        After or at the last recorded ionic step.

        """

        return self.last_stress

    @property
    def stress(self) -> np.ndarray | None:
        """
        Fetch stress.

        This container should contain all relevant stress.
        Currently, it only contains the final stress, which can be obtain
        by the id `final_stress`.

        """

        return self.final_stress

    @property
    def trajectory(self) -> dict[str, Any] | None:
        """
        Fetch unitcells, positions, species, forces and stress.

        For all calculation steps.

        """

        unitcell = self._content_parser.get_unitcell('all')
        positions = self._content_parser.get_positions('all')
        species = self._content_parser.get_species()
        forces = self._content_parser.get_forces('all')
        stress = self._content_parser.get_stress('all')
        # make sure all are sorted, first to last calculation
        # (species is constant)
        unitcell = sorted(unitcell.items())
        positions = sorted(positions.items())
        forces = sorted(forces.items())
        stress = sorted(stress.items())
        # convert to numpy
        unitcell = np.asarray([item[1] for item in unitcell])
        positions = np.asarray([item[1] for item in positions])
        forces = np.asarray([item[1] for item in forces])
        stress = np.asarray([item[1] for item in stress])
        # Aiida wants the species as symbols, so invert
        elements = _invert_dict(parsevaspct.elements)
        symbols = np.asarray([elements[item].title() for item in species.tolist()])

        if (
            (unitcell is not None)
            and (positions is not None)
            and (species is not None)
            and (forces is not None)
            and (stress is not None)
        ):
            trajectory_data = {}

            keys = ('cells', 'positions', 'symbols', 'forces', 'stress', 'steps')
            stepids = np.arange(unitcell.shape[0])

            for key, data in zip(keys, (unitcell, positions, symbols, forces, stress, stepids)):
                trajectory_data[key] = data
            return trajectory_data

        return None

    @property
    def total_energies(self) -> dict[str, float] | None:
        """Fetch the total energies after the last ionic run."""
        energies = self.energies
        if energies is None:
            return None
        energies_dict = {}
        for etype in self._settings.get('energy_type', self.DEFAULT_SETTINGS['energy_type']):
            energies_dict[etype] = energies[etype][-1]
            # Also return the raw electronic steps energy
            energies_dict[etype + '_electronic'] = energies[etype + '_electronic'][-1]

        return energies_dict

    @property
    def energies(self) -> dict[str, Any] | None:
        """Fetch the total energies."""
        # Check if we want total energy entries for each electronic step.
        electronic_step_energies = self._settings.get(
            'electronic_step_energies', self.DEFAULT_SETTINGS['electronic_step_energies']
        )

        return self._energies(nosc=not electronic_step_energies)

    def _energies(self, nosc: bool) -> dict[str, Any] | None:
        """
        Fetch the total energies for all energy types, calculations (ionic steps) and electronic steps.

        The returned dict from the parser contains the total energy types as a key (plus the _final, which is
        the final total energy ejected by VASP after the closure of the electronic steps). The energies can then
        be found in the flattened ndarray where the key `electronic_steps` indicate how many electronic steps
        there is per ionic step. Using the combination, one can rebuild the electronic step energy per ionic step etc.

        Because the VASPrun parser returns both the electronic step energies (at the end of each cycles) and the ionic
        step energies (_final), we apply a mapping to recovery the naming such that the ionic step energies do not have
        the suffix,
        but the electronic step energies do.
        """

        etype = self._settings.get('energy_type', self.DEFAULT_SETTINGS['energy_type'])
        # Create a copy
        etype = list(etype)
        etype_orig = list(etype)

        # Apply mapping and request the correct energies from the parsing results
        # VASP 5 has a bug where the energy_no_entropy is not included in the XML output - we have to calculate it here
        if self.version.startswith('5'):
            # For energy_no_entropy needs to be calculated here
            if 'energy_no_entropy' in etype_orig:
                etype.append('energy_free')
                etype.append('energy_extrapolated')

            # energy extrapolated is stored as energy_no_entropy for the ionic steps
            if 'energy_extrapolated' in etype_orig:
                etype.append('energy_no_entropy')

            # Remove duplicates
            etype = list(set(etype))
            energies = self._content_parser.get_energies(status='all', etype=etype, nosc=nosc)
            # Here we must calculate the true `energy_no_entropy`
            if 'energy_no_entropy' in etype_orig:
                # The energy_extrapolated_final is the entropy term itself in VASP 5
                # Store the calculated energy_no_entropy under 'energy_extrapolated_final',
                # which is then recovered as `energy_no_entropy` later
                energies['energy_extrapolated_final'] = (
                    energies['energy_free_final'] - energies['energy_extrapolated_final']
                )
        else:
            energies = self._content_parser.get_energies(status='all', etype=etype, nosc=nosc)

        if energies is None:
            return None

        # Apply mapping - those with `_final` has the suffix removed and those without has `_electronic` added
        mapped_energies = {}
        mapping = self.ENERGY_MAPPING_VASP5 if self.version.startswith('5') else self.ENERGY_MAPPING
        # Reverse the mapping - now key is the name of the original energies output
        revmapping = {value: key for key, value in mapping.items()}
        for key, value in energies.items():
            # Apply mapping if needed
            if key in revmapping:
                if revmapping[key].replace('_electronic', '') in etype_orig:
                    mapped_energies[revmapping[key]] = value
            else:
                mapped_energies[key] = value

        return mapped_energies

    @property
    def projectors(self) -> np.ndarray | None:
        """Fetch the projectors."""

        proj = self._content_parser.get_projectors()
        if proj is None:
            return None
        projectors = None
        prj = []
        try:
            prj.append(proj['total'])  # pylint: disable=unsubscriptable-object
        except KeyError:
            try:
                prj.append(proj['up'])  # pylint: disable=unsubscriptable-object
                prj.append(proj['down'])  # pylint: disable=unsubscriptable-object
            except KeyError:
                self._logger.error('Did not detect any projectors. Returning.')
        if len(prj) == 1:
            projectors = prj[0]
        else:
            projectors = np.asarray(prj)

        return projectors

    @property
    def dielectrics(self) -> dict[str, Any] | None:
        """Fetch the dielectric function."""

        diel = self._content_parser.get_dielectrics()
        if diel is None:
            return None
        dielectrics = {}
        energy = diel.get('energy')
        idiel = diel.get('imag')
        rdiel = diel.get('real')
        epsilon = diel.get('epsilon')
        epsilon_ion = diel.get('epsilon_ion')
        if energy is not None:
            dielectrics['ediel'] = energy
        if idiel is not None:
            dielectrics['rdiel'] = rdiel
        if rdiel is not None:
            dielectrics['idiel'] = idiel
        if epsilon is not None:
            dielectrics['epsilon'] = epsilon
        if epsilon_ion is not None:
            dielectrics['epsilon_ion'] = epsilon_ion

        return dielectrics

    @property
    def born_charges(self) -> np.ndarray | None:
        """Fetch the Born effective charges."""

        return self._content_parser.get_born()

    @property
    def hessian(self) -> np.ndarray | None:
        """Fetch the Hessian matrix."""

        hessian = self._content_parser.get_hessian()
        if hessian is None:
            return None
        return hessian

    @property
    def dynmat(self) -> dict[str, Any] | None:
        """Fetch the dynamical eigenvectors and eigenvalues."""

        dynmat = self._content_parser.get_dynmat()
        if dynmat is None:
            return None
        dyn = {}
        dyn['dynvec'] = dynmat['eigenvectors']  # pylint: disable=unsubscriptable-object
        dyn['dyneig'] = dynmat['eigenvalues']  # pylint: disable=unsubscriptable-object
        return dyn

    @property
    def dos(self) -> dict[str, Any] | None:
        """Fetch the total density of states."""

        dos = self._content_parser.get_dos()
        if dos is None:
            return None
        densta = {}
        # energy is always there, regardless of
        # total, spin or partial
        energy = dos['total']['energy']  # pylint: disable=unsubscriptable-object
        densta['energy'] = energy
        tdos = None
        pdos = None
        upspin = dos.get('up')
        downspin = dos.get('down')
        total = dos.get('total')
        if (upspin is not None) and (downspin is not None):
            tdos = np.stack((upspin['total'], downspin['total']))
            if (upspin['partial'] is not None) and (downspin['partial'] is not None):
                pdos = np.stack((upspin['partial'], downspin['partial']))
        else:
            tdos = total['total']
            pdos = total['partial']
        densta['tdos'] = tdos
        if pdos is not None:
            densta['pdos'] = pdos

        return densta

    @property
    def fermi_level(self) -> float | None:
        """Fetch Fermi level."""

        try:
            fermi_level = self._content_parser.get_fermi_level()
        except (ValueError, TypeError):
            fermi_level = None
        return fermi_level

    @property
    def run_status(self) -> dict[str, bool | None]:
        """Fetch run_status information"""
        info = {}
        # First check electronic convergence by comparing executed steps to the
        # maximum allowed number of steps (NELM).
        energies = self._content_parser.get_energies('last', nosc=False)
        parameters = self._content_parser.get_parameters()
        info['finished'] = not self._content_parser.truncated
        # Only set to true for untruncated run to avoid false positives
        if energies is None:
            info['electronic_converged'] = False
        elif energies.get('electronic_steps')[0] < parameters['nelm'] and not self._content_parser.truncated:
            info['electronic_converged'] = True
        else:
            info['electronic_converged'] = False

        # Then check the ionic convergence by comparing executed steps to the
        # maximum allowed number of steps (NSW).
        energies = self._content_parser.get_energies('all', nosc=True)
        if energies is None:
            info['ionic_converged'] = False
        elif len(energies.get('electronic_steps')) < parameters['nsw'] and not self._content_parser.truncated:
            info['ionic_converged'] = True
        else:
            info['ionic_converged'] = False
        # Override if nsw is 0 - no ionic steps are performed
        if parameters['nsw'] < 1:
            info['ionic_converged'] = None

        return info

    @property
    def band_properties(self) -> dict[str, Any] | None:
        """Fetch key properties of the electronic structure."""

        eigenvalues = self.eigenvalues
        occupancies = self.occupancies
        if eigenvalues is None:
            return None

        # Convert dict to index in numpy array
        if 'total' in eigenvalues:
            eig = np.array(eigenvalues['total'])
            occ = np.array(occupancies['total'])
        else:
            eig = np.array([eigenvalues['up'], eigenvalues['down']])
            occ = np.array([occupancies['up'], occupancies['down']])

        return get_band_properties(eig, occ)

    @property
    def parameters(self) -> dict[str, Any] | None:
        """
        Fetch the parsed input parameters which will include default values
        defined in the XML file.
        """

        parameters = self._content_parser.get_parameters()

        return parameters


def _build_structure(lattice: dict[str, Any]) -> dict[str, Any]:
    """Builds a structure according to AiiDA spec."""
    structure_dict = {}
    structure_dict['unitcell'] = lattice['unitcell']
    structure_dict['sites'] = []

    # AiiDA wants the species as symbols, so invert
    elements = _invert_dict(parsevaspct.elements)
    for pos, specie in zip(lattice['positions'], lattice['species']):
        site = {}
        site['position'] = np.dot(pos, lattice['unitcell'])
        site['symbol'] = elements[specie].title()
        site['kind_name'] = elements[specie].title()
        structure_dict['sites'].append(site)

    return structure_dict


def _invert_dict(dct: dict[Any, Any]) -> dict[Any, Any]:
    return dct.__class__(map(reversed, dct.items()))



"""
VASP Calculation Monitoring Functions.

This module provides monitoring functions for detecting and handling problems that may occur
during VASP calculations running on remote machines. These monitors are designed to identify
issues early and take preventive actions to avoid system crashes or resource waste.

The monitoring functions can detect:

* Stdout file overflow that could crash the AiiDA daemon
* Electronic loop timing issues that indicate inefficient calculations
* Stalled calculations that are no longer making progress

These monitors are typically used by AiiDA's calculation monitoring system to provide
real-time feedback about running VASP calculations and automatically handle problematic
situations.

.. note::
   These monitors operate on the remote machine where the VASP calculation is running
   and require a transport connection to access files and execute commands.
"""

import time
from pathlib import Path

from aiida.orm import CalcJobNode
from aiida.transports import Transport


def monitor_stdout(node: CalcJobNode, transport: Transport, size_threshold_mb: float = 5) -> str | None:
    """
    Monitor stdout file size to prevent overflow crashes.

    This function monitors the size of the VASP stdout file during calculation execution.
    If the file becomes too large, it indicates a potential problem (such as excessive
    output from convergence issues) that could crash the AiiDA daemon when attempting
    to retrieve and parse the file.

    When an oversized stdout file is detected, the function automatically truncates it
    to prevent system crashes, though this means the calculation is considered lost.

    :param node: The CalcJobNode representing the running VASP calculation
    :type node: CalcJobNode
    :param transport: Transport connection to the remote machine where VASP is running
    :type transport: Transport
    :param size_threshold_mb: Maximum allowed stdout file size in megabytes before
                             truncation occurs
    :type size_threshold_mb: int
    :returns: None if no overflow detected, otherwise an error message describing
              the overflow condition
    :rtype: str or None

    .. warning::
       When stdout overflow is detected, the calculation is automatically terminated
       by truncating the output file. This prevents system crashes but results in
       loss of the calculation.

    .. note::
       The default threshold of 5 MB is typically sufficient for normal VASP
       calculations. Larger thresholds may be needed for very large systems or
       calculations with verbose output.
    """

    # Check the current size of the VASP stdout file
    stdout_path = str(Path(node.get_remote_workdir()) / node.process_class._VASP_OUTPUT)
    file_stat = transport.get_attribute(stdout_path)
    if file_stat.st_size > 1024 * 1024 * size_threshold_mb:
        # Stdout file is dangerously large - truncate it to prevent system crashes
        # This typically indicates convergence problems or infinite loops in VASP
        # The calculation is lost, but we prevent the AiiDA daemon from crashing
        # when attempting to retrieve and parse the oversized file
        transport.exec_command_wait(f'truncate -s {size_threshold_mb}M {stdout_path}')
        return f'Very large stdout detected: {file_stat.st_size / (1024 * 1024):.2f} MB, potential critical crash.'
    # TODO: add detection of critical messages emitted when vasp got stuck


def monitor_loop_time(
    node: CalcJobNode,
    transport: Transport,
    minimum_electronic_loops: int = 10,
    patience_num_loops: int = 5,
    patience_minimum_time: float = 1800,
) -> str | None:
    """
    Monitor electronic loop timing to detect inefficient or stalled calculations.

    This function analyzes the timing of electronic self-consistency loops in VASP
    calculations to identify potential problems:

    1. **Slow convergence**: If electronic loops take too long relative to the walltime
       limit, the calculation may not complete within the allocated time.

    2. **Stalled calculations**: If the stdout file hasn't been updated for an
       extended period, the calculation may have crashed or become stuck.

    The function examines the OUTCAR file to extract loop timing information and
    compares it against the walltime limits and recent file modification times.

    :param node: The CalcJobNode representing the running VASP calculation
    :type node: CalcJobNode
    :param transport: Transport connection to the remote machine where VASP is running
    :type transport: Transport
    :param minimum_electronic_loops: Minimum number of electronic loops that should
                                   be completable within the walltime limit
    :type minimum_electronic_loops: int
    :param patience_num_loops: Number of loop times to wait before considering
                             a calculation stalled
    :type patience_num_loops: int
    :param patience_minimum_time: Minimum time in seconds to wait before checking
                                for stalled calculations
    :type patience_minimum_time: int
    :returns: None if timing is acceptable, otherwise an error message describing
              the detected problem (slow loops or stalled calculation)
    :rtype: str or None
    """

    outcar_path = str(Path(node.get_remote_workdir(), 'OURCAR'))
    stdout_path = str(Path(node.get_remote_workdir()) / node.process_class._VASP_OUTPUT)
    walltime_limit = node.get_option('max_wallclock_seconds')
    if walltime_limit is None:
        # Cannot monitor timing without a walltime limit
        return None

    # Extract electronic loop timings from OUTCAR file
    # LOOP entries contain timing information for each self-consistency cycle
    returncode, stdout, _ = transport.exec_command_wait(f"grep 'LOOP:' {outcar_path}")

    # Skip monitoring if no LOOP entries found (calculation hasn't started electronic steps)
    if returncode != 0:
        return None

    # Parse the last loop time from the grep output
    # Each LOOP line format: "LOOP:  CPU time  real time  (sec)  real time"
    last_loop_time = 0
    for line in stdout.splitlines():
        # Extract the real time (4th column) from each LOOP line
        # Keep updating to get the most recent loop time
        last_loop_time = float(line.strip().split()[-1])

    # Check if electronic loops are too slow relative to walltime
    # If each loop takes more than walltime/minimum_loops, we won't finish in time

    # Check for the presence of NELM
    incar = {key.lower(): value for key, value in node.inputs.get('parameters', {}).items()}
    nelm = incar.get('nelm', 60)
    # Remove none-scf loops for hybrid calculation
    if 'lhfcalc' in incar:
        nelm -= abs(incar.get('nelmdl', 5))

    # Take the minimum of the nelm and supplied minimum_electronic_loops
    # This is useful to avoid killing benchmarking calculations a low NELM setting
    minimum_electronic_loops = max(1, min(minimum_electronic_loops, nelm))

    if last_loop_time > walltime_limit / minimum_electronic_loops:
        return (
            f'Less than {minimum_electronic_loops} electronic loop can be run in this calculation due to '
            f'long electronic loop time: {last_loop_time:.2f} seconds.'
        )

    # Monitor for stalled calculations by checking stdout file modification time
    file_stat = transport.get_attribute(stdout_path)
    elapsed = time.time() - file_stat.st_mtime
    # Consider calculation stalled if:
    # 1. Minimum patience time has elapsed AND
    # 2. Time since last update exceeds expected time for several loops
    if elapsed > patience_minimum_time and elapsed > last_loop_time * patience_num_loops:
        return (
            f'Last update of the {stdout_path} file is more than {last_loop_time * patience_num_loops:.2f} '
            'seconds ago. It is very likely that the calculation is stalled or crashed.'
        )



import pytest
from aiida import orm
from ase.build import bulk

from aiida_vasp.workchains.v2.common import builder_updater as bup


def test_vasp_builder_updater(aiida_profile_clean, vasp_code, upload_potcar, potcar_family_name):
    structure = orm.StructureData(ase=bulk('MgO', 'rocksalt', 5.0)).store()
    vasp_code.store()
    upd = bup.VaspBuilderUpdater()
    upd.apply_preset(structure, code='vasp@localhost')
    assert upd.builder.structure == structure
    assert upd.builder.parameters['incar']['algo'] == 'normal'
    assert upd.builder.options['resources']['tot_num_mpiprocs'] == 1
    assert upd.builder.code == vasp_code
    assert upd.builder.settings.get_dict() == {}
    assert upd.builder.kpoints_spacing.value == 0.05
    assert upd.builder.potential_mapping.get_dict() == {'Mg': 'Mg_pv', 'O': 'O'}
    # The default family is not uploaded - this should remain None
    assert upd.builder.potential_family is None

    upd.set_potential_family(potcar_family_name)
    assert upd.builder.potential_family.value == potcar_family_name


def test_vasp_relax_updater(aiida_profile_clean, vasp_code):
    structure = orm.StructureData(ase=bulk('MgO', 'rocksalt', 5.0)).store()
    vasp_code.store()
    upd = bup.VaspRelaxUpdater()
    upd.apply_preset(structure, code='vasp@localhost')
    assert upd.builder.structure == structure
    assert upd.builder.vasp.parameters['incar']['algo'] == 'normal'
    assert upd.builder.vasp.options['resources']['tot_num_mpiprocs'] == 1
    assert upd.builder.vasp.code == vasp_code
    assert upd.builder.vasp.settings.get_dict() == {}
    assert upd.builder.vasp.kpoints_spacing.value == 0.05
    assert upd.builder.vasp.potential_family is None
    assert upd.builder.vasp.potential_mapping.get_dict() == {'Mg': 'Mg_pv', 'O': 'O'}
    assert upd.builder.relax_settings.get_dict()['algo'] == 'cg'


@pytest.mark.parametrize('hybrid', [True, False])
def test_vasp_band_updater(aiida_profile_clean, vasp_code, hybrid):
    structure = orm.StructureData(ase=bulk('MgO', 'rocksalt', 5.0)).store()
    vasp_code.store()
    if hybrid:
        upd = bup.VaspHybridBandUpdater()
    else:
        upd = bup.VaspBandUpdater()
    upd.apply_preset(structure, code='vasp@localhost')
    assert upd.builder.structure == structure
    assert upd.builder.scf.parameters['incar']['algo'] == 'normal'
    assert upd.builder.scf.options['resources']['tot_num_mpiprocs'] == 1
    assert upd.builder.scf.code == vasp_code
    assert upd.builder.scf.settings.get_dict() == {}
    assert upd.builder.scf.kpoints_spacing.value == 0.05
    assert upd.builder.scf.potential_family is None
    assert upd.builder.scf.potential_mapping.get_dict() == {'Mg': 'Mg_pv', 'O': 'O'}

    if hybrid:
        upd = bup.VaspHybridBandUpdater()
    else:
        upd = bup.VaspBandUpdater()
    upd.apply_preset(structure, code='vasp@localhost', run_relax=True)

    assert upd.builder.structure == structure
    assert upd.builder.relax.vasp.parameters['incar']['algo'] == 'normal'
    assert upd.builder.relax.vasp.options['resources']['tot_num_mpiprocs'] == 1
    assert upd.builder.relax.vasp.code == vasp_code
    assert upd.builder.relax.vasp.settings.get_dict() == {}
    assert upd.builder.relax.vasp.kpoints_spacing.value == 0.05
    assert upd.builder.relax.vasp.potential_family is None
    assert upd.builder.relax.vasp.potential_mapping.get_dict() == {'Mg': 'Mg_pv', 'O': 'O'}


def test_vasp_neb_updater(aiida_profile_clean, vasp_code):
    structure = orm.StructureData(ase=bulk('MgO', 'rocksalt', 5.0)).store()
    vasp_code.store()
    upd = bup.VaspNEBUpdater()
    upd.apply_preset(structure, structure, code='vasp@localhost')



"""Unittests for VaspCalculation."""

# pylint: disable=unused-import,redefined-outer-name,unused-argument,unused-wildcard-import,wildcard-import, import-outside-toplevel
import contextlib
import os
import tempfile

import pytest
from aiida import orm
from aiida.common import ValidationError

from aiida_vasp.calcs.vasp import VaspCalculation
from aiida_vasp.parsers.content_parsers.potcar import MultiPotcarIo


@pytest.mark.parametrize(['vasp_structure', 'vasp_kpoints'], [('cif', 'mesh')], indirect=True)
def test_write_incar(vasp_calc_and_ref):
    """Write parameters input node to INCAR object, compare to reference string."""
    vasp_calc, reference = vasp_calc_and_ref
    with managed_temp_object() as temp_object:
        vasp_calc.write_incar(temp_object)
        with open(temp_object, 'r', encoding='utf8') as result_incar_fo:
            assert result_incar_fo.readlines() == reference['incar']


@pytest.mark.parametrize(['vasp_structure', 'vasp_kpoints'], [('cif', 'mesh')], indirect=True)
def test_write_potcar(vasp_calc_and_ref):
    """Check that POTCAR is written correctly."""
    vasp_calc, _ = vasp_calc_and_ref

    with managed_temp_object() as temp_object:
        vasp_calc.write_potcar(temp_object)
        with open(temp_object, 'r', encoding='utf8') as potcar_fo:
            result_potcar = potcar_fo.read()
        assert 'In_sv' in result_potcar
        assert 'As' in result_potcar
        assert 'In_d' in result_potcar
        assert result_potcar.count('End of Dataset') == 2, result_potcar

        if isinstance(vasp_calc.inputs.structure, orm.StructureData):
            multipotcar = MultiPotcarIo.read(temp_object)
            potcar_order = [potcar.node.full_name for potcar in multipotcar.potcars]
            assert potcar_order == ['In_sv', 'As', 'In_d', 'As']


@pytest.mark.parametrize(['vasp_structure', 'vasp_kpoints'], [('cif', 'mesh')], indirect=True)
def test_write_chgcar(vasp_calc, vasp_inputs, vasp_chgcar, sandbox_folder):
    """Test that CHGAR object is written correctly."""
    chgcar, _ = vasp_chgcar

    inputs = vasp_inputs(
        parameters={
            'gga': 'PE',
            'gga_compat': False,
            'lorbit': 11,
            'sigma': 0.5,
            'magmom': '30 * 2*0.',
            'icharg': 1,
        }
    )

    inputs.charge_density = chgcar
    calc = vasp_calc(inputs=inputs)

    calcinfo = calc.prepare_for_submission(sandbox_folder)
    assert 'CHGCAR' in [item[1] for item in calcinfo.local_copy_list]


@pytest.mark.parametrize(['vasp_structure', 'vasp_kpoints'], [('cif', 'mesh')], indirect=True)
def test_write_wavecar(vasp_calc, vasp_inputs, vasp_wavecar, sandbox_folder):
    """Test that WAVECAR object is written correctly."""
    wavecar, _ = vasp_wavecar
    inputs = vasp_inputs(
        parameters={
            'gga': 'PE',
            'gga_compat': False,
            'lorbit': 11,
            'sigma': 0.5,
            'magmom': '30 * 2*0.',
            'istart': 1,
        }
    )
    inputs.wavefunctions = wavecar
    calc = vasp_calc(inputs=inputs)
    calcinfo = calc.prepare_for_submission(sandbox_folder)

    assert 'WAVECAR' in [item[1] for item in calcinfo.local_copy_list]


@pytest.mark.parametrize(['vasp_structure', 'vasp_kpoints'], [('cif', 'mesh')], indirect=True)
def test_incar_validate(vasp_calc, vasp_inputs, sandbox_folder):
    """Test incar with invalid tags raises exception."""
    inputs_dict = {
        'gga': 'PE',
        'smear': 3,  # <- Invalid tag
    }
    inputs = vasp_inputs(parameters=inputs_dict)
    calc = vasp_calc(inputs=inputs)

    with pytest.raises(ValidationError):
        calc.prepare_for_submission(sandbox_folder)

    inputs = vasp_inputs(parameters=inputs_dict, settings={'skip_parameters_validation': True})
    calc = vasp_calc(inputs=inputs)
    calc.prepare_for_submission(sandbox_folder)

    inputs = vasp_inputs(parameters=inputs_dict, settings={'unsupported_parameters': {'smear': {}}})
    calc = vasp_calc(inputs=inputs)
    calc.prepare_for_submission(sandbox_folder)


@pytest.mark.parametrize(['vasp_structure', 'vasp_kpoints'], [('cif', 'mesh')], indirect=True)
def test_prepare(vasp_calc, vasp_chgcar, vasp_wavecar, vasp_inputs, sandbox_folder):
    """Check that preparing creates all necessary objects."""
    wavecar, _ = vasp_wavecar
    chgcar, _ = vasp_chgcar

    inputs_dict = {
        'gga': 'PE',
        'gga_compat': False,
        'lorbit': 11,
        'sigma': 0.5,
        'magmom': '30 * 2*0.',
        'icharg': 11,
    }

    inputs = vasp_inputs(parameters=inputs_dict)
    inputs.charge_density = chgcar
    inputs.wavefunctions = wavecar

    calc = vasp_calc(inputs=inputs)
    temp_folder = sandbox_folder
    calcinfo = calc.prepare_for_submission(temp_folder)
    input_objects = temp_folder.get_content_list()

    for name in ['INCAR', 'KPOINTS', 'POSCAR', 'POTCAR']:
        assert name in input_objects

    assert calcinfo.codes_info[0].stdout_name == VaspCalculation._VASP_OUTPUT
    assert calcinfo.codes_info[0].join_files is True

    inputs_dict.update({'icharg': 2})

    inputs = vasp_inputs(parameters=inputs_dict)
    inputs.charge_density = chgcar
    inputs.wavefunctions = wavecar

    calc = vasp_calc(inputs=inputs)
    calcinfo = calc.prepare_for_submission(sandbox_folder)

    assert 'WAVECAR' in [item[1] for item in calcinfo.local_copy_list]


@pytest.mark.parametrize(['vasp_structure', 'vasp_kpoints'], [('cif', 'mesh')], indirect=True)
def test_verify_success(vasp_calc_and_ref):
    """Check that correct inputs are successfully verified."""
    vasp_calc, _ = vasp_calc_and_ref
    vasp_calc.verify_inputs()


@pytest.mark.parametrize(['vasp_structure', 'vasp_kpoints'], [('cif', 'mesh')], indirect=True)
def test_verify_fail(vasp_calc, vasp_inputs):
    """Check that incorrect inputs are not verified."""
    inputs = vasp_inputs()
    inputs.pop('kpoints')

    with pytest.raises(ValueError):
        vasp_calc(inputs=inputs)


@contextlib.contextmanager
def managed_temp_object():
    """Create a temp file object for a with context, delete after use."""
    _, temp_object = tempfile.mkstemp()
    try:
        yield temp_object
    finally:
        os.remove(temp_object)


@pytest.mark.parametrize(['vasp_structure', 'vasp_kpoints'], [('str', 'mesh')], indirect=True)
def test_vasp_calc_only_output(run_vasp_process):
    """Test a run of a basic VASP calculation which misses critical files."""
    _, node = run_vasp_process(test_case='stdout')
    assert node.exit_status == 352
    vasp_output = node.outputs.retrieved.get_object_content('vasp_output')
    assert 'MOCK PREPEND: Using test data from folder: stdout' in vasp_output
    assert 'VERY BAD NEWS! internal error in subroutine IBZKPT:' in vasp_output


@pytest.mark.parametrize(['vasp_structure', 'vasp_kpoints'], [('str', 'mesh')], indirect=True)
def test_vasp_calc(run_vasp_process):
    """Test a run of a basic VASP calculation and its details."""

    results, node = run_vasp_process(settings={'parser_settings': {'check_errors': False}})
    assert node.exit_status == 0

    # Check that the standard output is there
    assert 'retrieved' in results
    assert 'misc' in results
    assert 'remote_folder' in results

    # Also, we need to check that there is some minimum content in misc, as it will contain
    # None if parsing somehow failed.
    misc = results['misc'].get_dict()
    assert 'total_energies' in misc
    assert 'stress' in misc
    assert 'run_status' in misc
    assert 'run_stats' in misc

    # By default we should store all always retrieve objects
    retrieve_temporary_list_ref = []
    retrieve_list_ref = VaspCalculation._ALWAYS_RETRIEVE_LIST + [
        '_scheduler-stdout.txt',
        '_scheduler-stderr.txt',
    ]
    retrieve_temporary_list = node.get_retrieve_temporary_list()
    retrieve_list = node.get_retrieve_list()
    assert retrieve_temporary_list == retrieve_temporary_list_ref
    assert set(retrieve_list_ref) == set(retrieve_list)
    objects = node.outputs.retrieved.base.repository.list_objects()
    names = [single_object.name for single_object in objects]
    # Exclude Wannier objects as they are not in the test set
    retrieve_list_ref_no_wannier = [item for item in retrieve_list_ref if 'wannier' not in item]
    assert set(names) == set(retrieve_list_ref_no_wannier)

    # Check that we always try to parse notifications
    assert misc.get('notifications') is not None

    # vasp_output contains boxed WARNING
    assert len(misc['notifications']) == 1
    assert misc['notifications'][0]['kind'] == 'WARNING'


@pytest.mark.parametrize(['vasp_structure', 'vasp_kpoints'], [('str', 'mesh')], indirect=True)
def test_vasp_calc_delete(run_vasp_process):
    """Test a run of a basic VASP calculation where one does not want to store
    the always retrieved objects after parsing.
    """
    retrieve_list_ref = ['_scheduler-stdout.txt', '_scheduler-stderr.txt']
    inputs = {}
    inputs['settings'] = orm.Dict(dict={'ALWAYS_STORE': False})
    _, node = run_vasp_process(inputs)
    objects = node.outputs.retrieved.base.repository.list_objects()
    names = [single_object.name for single_object in objects]
    assert set(names) == set(retrieve_list_ref)


@pytest.mark.parametrize(['vasp_structure', 'vasp_kpoints'], [('str', 'mesh')], indirect=True)
def test_vasp_calc_extra(run_vasp_process):
    """Test a run of a basic VASP calculation where one wants to keep additional objects after parsing is completed."""
    # Let us add an additional object to the retrieve_list (which do not delete the object after parse)
    # and check if it is actually there
    inputs = {}
    extra_object_to_keep = 'POSCAR'
    inputs['settings'] = orm.Dict(dict={'ADDITIONAL_RETRIEVE_LIST': [extra_object_to_keep]})
    _, node = run_vasp_process(inputs)
    retrieve_temporary_list_ref = []
    retrieve_list_ref = VaspCalculation._ALWAYS_RETRIEVE_LIST + [
        '_scheduler-stdout.txt',
        '_scheduler-stderr.txt',
        'POSCAR',
    ]
    retrieve_temporary_list = node.get_retrieve_temporary_list()
    retrieve_list = node.get_retrieve_list()
    assert retrieve_temporary_list == retrieve_temporary_list_ref
    assert set(retrieve_list_ref) == set(retrieve_list)
    objects = node.outputs.retrieved.base.repository.list_objects()
    names = [single_object.name for single_object in objects]
    # Exclude Wannier objects as they are not in the test set
    retrieve_list_ref_no_wannier = [item for item in retrieve_list_ref if 'wannier' not in item]
    assert set(names) == set(retrieve_list_ref_no_wannier)


@pytest.mark.parametrize(['vasp_structure', 'vasp_kpoints'], [('str', 'mesh')], indirect=True)
def test_vasp_calc_delete_extra(run_vasp_process):
    """Test a run of a basic VASP calculation where one wants to retrieve additional
    objects but not store them after parsing."""
    # Let us add an additional object to the retrieve_list (which do not delete the object after parse)
    # and check if it is actually there
    retrieve_list_ref = ['_scheduler-stdout.txt', '_scheduler-stderr.txt']
    inputs = {}
    extra_object_to_keep = 'POSCAR'
    inputs['settings'] = orm.Dict(
        dict={
            'ALWAYS_STORE': False,
            'ADDITIONAL_RETRIEVE_TEMPORARY_LIST': [extra_object_to_keep],
        },
    )
    _, node = run_vasp_process(inputs)
    retrieve_list_ref = ['_scheduler-stdout.txt', '_scheduler-stderr.txt']
    retrieve_temporary_list_ref = VaspCalculation._ALWAYS_RETRIEVE_LIST + ['POSCAR']
    retrieve_list = node.get_retrieve_list()
    retrieve_temporary_list = node.get_retrieve_temporary_list()
    assert set(retrieve_temporary_list) == set(retrieve_temporary_list_ref)
    assert set(retrieve_list) == set(retrieve_list_ref)
    objects = node.outputs.retrieved.base.repository.list_objects()
    names = [single_object.name for single_object in objects]
    assert set(names) == set(retrieve_list_ref)


@pytest.mark.parametrize(['vasp_structure', 'vasp_kpoints'], [('str', 'mesh')], indirect=True)
def test_vasp_calc_del_str_ext(run_vasp_process):
    """Test a run of a basic VASP calculation where one wants to retrieve additional objects and store only those."""
    # Let us add an additional object to the retrieve_list (which do not delete the object after parse)
    # and check if it is actually there
    retrieve_list_ref = ['_scheduler-stdout.txt', '_scheduler-stderr.txt']
    inputs = {}
    extra_object_to_keep = 'POSCAR'
    inputs['settings'] = orm.Dict(
        dict={
            'ALWAYS_STORE': False,
            'ADDITIONAL_RETRIEVE_LIST': [extra_object_to_keep],
        },
    )
    _, node = run_vasp_process(inputs)
    retrieve_list_ref = ['_scheduler-stdout.txt', '_scheduler-stderr.txt', 'POSCAR']
    retrieve_temporary_list_ref = VaspCalculation._ALWAYS_RETRIEVE_LIST
    retrieve_list = node.get_retrieve_list()
    retrieve_temporary_list = node.get_retrieve_temporary_list()
    assert set(retrieve_temporary_list) == set(retrieve_temporary_list_ref)
    assert set(retrieve_list_ref) == set(retrieve_list)
    retrieve_list_ref = ['_scheduler-stdout.txt', '_scheduler-stderr.txt', 'POSCAR']
    objects = node.outputs.retrieved.base.repository.list_objects()
    names = [single_object.name for single_object in objects]
    assert set(names) == set(retrieve_list_ref)


@pytest.mark.parametrize(['vasp_structure', 'vasp_kpoints'], [('str', 'mesh')], indirect=True)
def test_vasp_no_potcar_in_repo(run_vasp_process):
    """Test a VASP run to verify that there is no POTCAR object in the repository."""
    # Let us add an additional object to the retrieve_list (which do not delete the object after parse)
    # and check if it is actually there
    inputs = {}
    _, node = run_vasp_process(inputs)
    repo_objects = node.base.repository.list_object_names()
    assert 'POTCAR' not in repo_objects


@pytest.mark.parametrize(
    'test_case,expected,has_notification',
    [
        ['exit_codes/converged-with-error', 703, True],
        ['exit_codes/converged', 0, False],
        ['exit_codes/unfinished', 700, False],
        ['exit_codes/elec-unconverged', 701, False],
        ['exit_codes/ionic-unconverged', 702, False],
    ],
)
@pytest.mark.parametrize(
    [
        'vasp_structure',
        'vasp_kpoints',
    ],
    [('str', 'mesh')],
    indirect=True,
)
def test_vasp_calc_exit_codes(run_vasp_process, test_case, expected, has_notification):
    """
    Test running a VASP calculation with electronic/ionic convergence problems and
    check if the exit_codes are set accordingly.
    """
    results, node = run_vasp_process(test_case=test_case)

    # Check that the standard output is there
    assert 'retrieved' in results
    assert 'misc' in results
    assert 'remote_folder' in results

    misc = results['misc'].get_dict()
    assert node.exit_status == expected
    assert bool(misc['notifications']) is has_notification


@pytest.mark.parametrize(
    [
        'vasp_structure',
        'vasp_kpoints',
    ],
    [('str', 'mesh')],
    indirect=True,
)
def test_vasp_calc_error_suppress(run_vasp_process):
    """
    Test running a VASP calculation with electronic/ionic convergence problems and
    check if the exit_codes are set accordingly.
    """
    results, node = run_vasp_process(
        test_case='exit_codes/converged-with-error',
        settings={
            'parser_settings': {'critical_notification_errors': [], 'check_errors': True},
        },
    )

    # Check that the standard output is there
    assert 'retrieved' in results
    assert 'misc' in results
    assert 'remote_folder' in results

    misc = results['misc'].get_dict()
    assert node.exit_status == 0
    assert bool(misc['notifications'])


@pytest.mark.parametrize(
    [
        'vasp_structure',
        'vasp_kpoints',
    ],
    [('str', 'mesh')],
    indirect=True,
)
def test_vasp_calc_error_ignore_all(run_vasp_process):
    """
    Test running a VASP calculation with electronic/ionic convergence problems and
    check if the exit_codes are set accordingly.
    """
    results, node = run_vasp_process(
        test_case='exit_codes/converged-with-error',
        settings={'parser_settings': {'ignore_notification_errors': True}},
    )

    # Check that the standard output is there
    assert 'retrieved' in results
    assert 'misc' in results
    assert 'remote_folder' in results

    misc = results['misc'].get_dict()
    assert node.exit_status == 0
    assert bool(misc['notifications'])



"""
Utils for AiiDA.

Utilities for making working against AiiDA a bit easier. Mostly here due to
historical reasons when AiiDA was rapidly developed. In the future most routines
that have now standardized in AiiDA will be removed.
"""

# ruff: noqa: PLC0415
from __future__ import annotations

import warnings
from functools import wraps
from typing import Any, Callable

import numpy as np
from aiida import __version__ as aiida_version_
from aiida import orm
from aiida.common.exceptions import MissingEntryPointError
from aiida.orm import AuthInfo, QueryBuilder, User, load_node
from aiida.plugins import DataFactory
from packaging import version

BASIC_DATA_TYPES: list[str] = ['core.bool', 'core.float', 'core.int', 'core.list', 'core.str', 'core.dict']


def querybuild(cls: type, **kwargs: Any) -> QueryBuilder:
    """
    Instantiates and returns a QueryBuilder instance.

    The QueryBuilder's path has one vertice so far, namely this class.
    Additional parameters (e.g. filters or a label),
    can be passes as keyword arguments.

    :param label: Label to give
    :param filters: filters to apply
    :param project: projections
    :returns: a QueryBuilder instance.
    """

    query_builder = QueryBuilder()
    filters = kwargs.pop('filters', {})
    query_builder.append(cls, filters=filters, **kwargs)

    return query_builder


def get_data_class(data_type: str) -> type:
    """Provide access to the orm.data classes with deferred dbenv loading."""

    data_cls = None
    try:
        data_cls = DataFactory(data_type)
    except MissingEntryPointError as err:
        raise err
    return data_cls


def get_current_user() -> User:
    """Get current user."""
    current_user = User.collection.get_default()
    return current_user


def copy_parameter(old_parameter: orm.Dict) -> orm.Dict:
    """Assemble a new Dict."""
    return orm.Dict(dict=old_parameter.get_dict())


def displaced_structure(structure: orm.StructureData, displacement: np.ndarray, entry: int) -> orm.StructureData:
    disp_structure = structure.clone()
    displace_position(disp_structure, displacement, entry)
    return disp_structure


def compressed_structure(structure: orm.StructureData, volume_change: float) -> orm.StructureData:
    comp_structure = structure.clone()
    compress_cell(comp_structure, volume_change)
    return comp_structure


def displace_position(structure: orm.StructureData, displacement: np.ndarray, entry: int) -> None:
    """Displace a position in the StructureData."""
    sites = structure.sites
    positions = []
    for site in sites:
        positions.append(site.position)
    new_position = np.asarray(positions[entry - 1]) + displacement
    new_position = new_position.tolist()
    positions[entry - 1] = tuple(new_position)
    structure.reset_sites_positions(positions)


def compress_cell(structure: orm.StructureData, volume_change: float) -> None:
    """Apply compression or tensile forces to the unit cell."""
    cell = structure.cell
    new_cell = np.array(cell) * volume_change
    structure.reset_cell(new_cell.tolist())


def aiida_version() -> version.Version:
    return version.parse(aiida_version_)


def cmp_version(string: str) -> version.Version:
    return version.parse(string)


def cmp_load_verdi_data() -> Any:
    """Load the verdi data click command group for any version since 0.11."""
    verdi_data = None
    import_errors = []

    try:
        from aiida.cmdline.commands import data_cmd as verdi_data
    except ImportError as err:
        import_errors.append(err)

    if not verdi_data:
        try:
            from aiida.cmdline.commands import verdi_data
        except ImportError as err:
            import_errors.append(err)

    if not verdi_data:
        try:
            from aiida.cmdline.commands.cmd_data import verdi_data
        except ImportError as err:
            import_errors.append(err)

    if not verdi_data:
        err_messages = '\n'.join([f' * {err}' for err in import_errors])
        raise ImportError('The verdi data base command group could not be found:\n' + err_messages)

    return verdi_data


def create_authinfo(computer: orm.Computer, store: bool = False) -> AuthInfo:
    """Allow the current user to use the given computer."""

    authinfo = AuthInfo(computer=computer, user=get_current_user())
    if store:
        authinfo.store()
    return authinfo


def cmp_get_authinfo(computer: orm.Computer) -> AuthInfo | None:
    """Get an existing authinfo or None for the given computer and current user."""
    return computer.get_authinfo(get_current_user())


def cmp_get_transport(computer: orm.Computer) -> Any:
    if hasattr(computer, 'get_transport'):
        return computer.get_transport()
    authinfo = cmp_get_authinfo(computer)
    return authinfo.get_transport()


def ensure_node_first_arg(func: Callable[..., Any]) -> Callable[..., Any]:
    """Decorator to load a node if it is passed as a string."""

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        """Make sure the first node is a Node instance."""
        if len(args) > 0:
            node = args[0]
            if not isinstance(node, orm.Node):
                node = load_node(node)
        args = list(args)
        args[0] = node
        return func(*args, **kwargs)

    return wrapper


def ensure_node_kwargs(func: Callable[..., Any]) -> Callable[..., Any]:
    """Decorator to load a node if it is passed as a key word argument ends with 'node'."""

    @wraps(func)
    def wrapper(node: Any, *args: Any, **kwargs: Any) -> Any:
        """Make sure the key world arguments ends with '_node' node is a Node instance."""
        new_kwargs = dict(kwargs)
        for name, value in kwargs.items():
            if name.endswith('node'):
                if not isinstance(value, orm.Node):
                    new_kwargs[name] = load_node(value)
        return func(node, *args, **new_kwargs)

    return wrapper


def convert_dict_case(
    dict_in: dict[str, Any],
    recursive: bool = True,
    warn: bool = False,
    lower: bool = True,
    raise_convert: bool = False,
) -> dict[str, Any]:
    """
    Recursively convert the keys of a dictionary to lower or upper cases, returns a new dictionary.

    :param dict_in: The input dictionary whose keys need to be converted.
    :param recursive: If True, the function will recursively convert keys in nested dictionaries.
    :param warn: If True, the function will print a warning if a key is converted.
    :param lower: If True, convert keys to lowercase; otherwise, convert to uppercase.
    :param raise_convert: If True, raise an error if a key is converted.
    :return: A new dictionary with keys converted to the specified case.
    """

    converted_dict = {}
    for key, value in dict_in.items():
        new_key = key.lower() if lower else key.upper()
        if new_key != key:
            expected = 'upper' if lower is False else 'lower'
            if warn:
                expected = 'upper' if lower is False else 'lower'
                warnings.warn(f"Key '{key}' converted to '{new_key}' - please use {expected} case keys")
            if raise_convert:
                raise ValueError(f"Key '{key}' converted to '{new_key}' - please use {expected} case keys")

        if recursive and isinstance(value, dict):
            converted_dict[new_key] = convert_dict_case(value, recursive, warn, lower, raise_convert)
        else:
            converted_dict[new_key] = value
    return converted_dict



"""
A general archive class.

Archive data class: store multiple files together in a compressed archive in the repository.
"""

from __future__ import annotations

import os
import tarfile
import tempfile
from contextlib import contextmanager
from typing import Any, Generator

from aiida.orm.nodes import Data


class ArchiveData(Data):
    """Compressed archive data node, contains a group of files that don't need to be readily accessible on their own."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._filelist: list[tuple[str, str]] = []
        super().__init__(*args, **kwargs)

    @contextmanager
    def get_archive(self) -> Generator[tarfile.TarFile, None, None]:
        with self.base.repository.open('archive.tar.gz', mode='rb') as fobj:  # pylint: disable=not-context-manager
            with tarfile.open(fileobj=fobj, mode='r:gz') as tar:
                yield tar

    @contextmanager
    def archive(self) -> Generator[tarfile.TarFile, None, None]:
        with self.base.repository.open('archive.tar.gz', mode='rb') as fobj:  # pylint: disable=not-context-manager
            with tarfile.open(fileobj=fobj, mode='r:gz') as tar:
                yield tar

    def get_archive_list(self) -> None:
        with self.get_archive() as archive:
            return archive.list()

    def add_file(self, src_abs: str, dst_filename: str | None = None) -> None:
        if not dst_filename:
            dst_filename = os.path.basename(src_abs)
        self._filelist.append((src_abs, dst_filename))

    def _make_archive(self) -> None:
        """Create the archive file on disk with all it's contents."""
        _, path = tempfile.mkstemp()
        try:
            with tarfile.open(path, mode='w:gz') as archive:
                for src, dstn in self._filelist:
                    archive.add(src, arcname=dstn)
            self.base.repository.put_object_from_file(path, path='archive.tar.gz')
        finally:
            os.remove(path)

    # pylint: disable=arguments-differ, signature-differs
    def store(self, *args: Any, **kwargs: Any) -> Data:
        self._make_archive()
        del self._filelist
        return super().store(*args, **kwargs)



"""Unittests for Vasp2w90Calculation."""

# pylint: disable=unused-import,redefined-outer-name,unused-argument,unused-wildcard-import,wildcard-import, import-outside-toplevel
from __future__ import absolute_import, print_function

import contextlib
import os
import re
import tempfile

import pytest


def normalize_contents(object_contents):
    """Remove trailing zeroes after floating point and normalize trailin newline to unix standard."""
    normalized = re.sub(r'(\d*.\d*?)0+(\s)', r'\g<1>0\2', object_contents)  # remove trailing zeroes
    if not re.match(r'\n', object_contents[-1]):  # add trailing newline if necessary
        normalized += '\n'
    return normalized


def assert_contents_equivalent(contents_a, contents_b):
    """Assert equivalence of objects with floating point numbers."""
    assert normalize_contents(contents_a) == normalize_contents(contents_b)


@pytest.mark.parametrize(['vasp_structure', 'vasp_kpoints'], [('cif', 'mesh')], indirect=True)
def test_wannier_parameters(vasp2w90_calc_and_ref):
    """Test that Wannier90 parameters and projections can be specified."""
    vasp_calc, _ = vasp2w90_calc_and_ref
    projections = vasp_calc.inputs.wannier_projections.get_list()
    parameters = vasp_calc.inputs.wannier_parameters.get_dict()
    assert projections == ['Ga : s; px; py; pz', 'As : px; py; pz']
    assert parameters == {'spinors': True, 'num_iter': 0, 'num_wann': 14, 'num_bands': 24, 'dis_num_iter': 1000}


@pytest.mark.parametrize(['vasp_structure', 'vasp_kpoints'], [('cif', 'mesh')], indirect=True)
def test_prepare_for_submission(vasp2w90_calc_and_ref, sandbox_folder):
    """Test that the lwannier90 flag is written at the prepare for submission step."""
    vasp_calc, reference = vasp2w90_calc_and_ref
    with pytest.raises(NotImplementedError):
        vasp_calc.prepare_for_submission(sandbox_folder)
        with managed_temp_object() as temp_object:
            vasp_calc.write_incar(temp_object)
            with open(temp_object, 'r', encoding='utf8') as result_incar_fo:
                assert result_incar_fo.readlines() == reference['incar']


@pytest.mark.parametrize(['vasp_structure', 'vasp_kpoints'], [('cif', 'mesh')], indirect=True)
def test_write_win(vasp2w90_calc_and_ref):
    """Write wannier90.win input object and compare to reference."""
    vasp_calc, reference = vasp2w90_calc_and_ref
    with managed_temp_object() as temp_object:
        with pytest.raises(NotImplementedError):
            vasp_calc.write_win(temp_object)
            with open(temp_object, 'r', encoding='utf8') as result_incar_fo:
                assert result_incar_fo.read() == reference['win']


@contextlib.contextmanager
def managed_temp_object():
    """Create a temp file object for a with context, delete after use."""
    _, temp_object = tempfile.mkstemp()
    try:
        yield temp_object
    finally:
        os.remove(temp_object)



"""
The ``POSCAR``/``CONTCAR`` parser interface.

Contains the parsing interfaces to ``parsevasp`` used to parse ``POSCAR``/``CONTCAR`` content.
"""

from typing import Any, TextIO

import numpy as np
from aiida import orm
from aiida.common.constants import elements
from parsevasp.poscar import Poscar, Site

from aiida_vasp.parsers.content_parsers.base import BaseFileParser


class PoscarParser(BaseFileParser):
    """The parser interface that enables parsing of ``POSCAR``/``CONTCAR`` content.

    The parser is triggered by using the ``poscar-structure`` quantity key. The quantity key ``structure``
    will on the other hand parse the structure using the XML parser.

    :param precision: An integer specifying the number of digits for floating point numbers that will be written
                      to ``POSCAR``/``CONTCAR``. Defaults to 12.
    :type precision: int, optional
    """

    DEFAULT_SETTINGS = {'quantities_to_parse': ['structure']}

    PARSABLE_QUANTITIES = {
        'structure': {
            'inputs': [],
            'name': 'structure',
            'prerequisites': [],
        },
    }

    def __init__(self, *, precision=12, **kwargs):
        """Initialize an instance of this class."""

        self._precision = precision
        super().__init__(**kwargs)

    def _init_from_handler(self, handler: TextIO) -> None:
        """Initialize using a file like handler.

        :param handler: A file like object that provides the necessary ``POSCAR``/``CONTCAR`` content to be parsed.
        :type handler: file-like object
        """

        try:
            self._content_parser = Poscar(
                file_handler=handler, prec=self._precision, conserve_order=True, logger=self._logger
            )
        except SystemExit:
            self._logger.warning('Parsevasp exited abnormally.')

    def _init_from_data(self, data: orm.StructureData) -> None:
        """Initialize using AiiDA ``StructureData`` instance.

        :param data: A valid AiiDA ``StructureData`` object.
        :type data: object
        """

        if isinstance(data, orm.StructureData):
            self._content_data = data
        else:
            raise TypeError('The supplied AiiDA data structure is not a StructureData.')

    @property
    def structure(self) -> dict[str, Any]:
        """Return structure from POSCAR.

        :returns: A dict that contain keys ``comment``, ``unitcell`` and ``sites``, which are compatible
                  with consumption of the initialization of the AiiDA ``StructureData``.
        :rtype: dict
        """

        aiida_structure = parsevasp_to_aiida(self._content_parser)

        return aiida_structure

    def _content_data_to_content_parser(self) -> 'PoscarParser':
        """Convert an AiiDA ``StructureData`` to a content parser instance of ``Poscar`` from ``parsevasp``.

        :returns: An instance of ``Poscar`` from ``parsevasp``.
        :rtype: object
        """
        dictionary = {}
        dictionary['comment'] = self._content_data.label or self._content_data.get_formula()
        dictionary['unitcell'] = np.asarray(self._content_data.cell)
        # As for now all Aiida-structures are in Cartesian coordinates.
        direct = False
        sites = []
        _transform_to_bool = np.vectorize(self.transform_to_bool)
        for index, site in enumerate(self._content_data.sites):
            if self._options is None:
                _selective = [True, True, True]
            else:
                try:
                    _selective = _transform_to_bool(np.array(self._options['positions_dof'])[index, :])
                except KeyError:
                    _selective = [True, True, True]
            sites.append(Site(site.kind_name, site.position, selective=_selective, direct=direct))

        dictionary['sites'] = sites

        # We brake hard if ``parsevasp`` fail here. If we can not write we will not try another parser.
        content_parser = Poscar(poscar_dict=dictionary, prec=self._precision, conserve_order=True, logger=self._logger)

        return content_parser

    def transform_to_bool(self, value: str | int) -> bool:
        """Helper function to transform the dictionary from strings or integers to bools"""
        if value in [0, 'F', 'f']:
            return False
        if value in [1, 'T', 't']:
            return True
        return True


def parsevasp_to_aiida(poscar: Poscar) -> dict[str, Any]:
    """``parsevasp`` to AiiDA conversion.

    Generate an AiiDA structure that can be consumed by ``StructureData`` on initialization
    from the ``parsevasp`` instance of the ``Poscar`` class.

    :param poscar: An instance of the ``Poscar`` class in ``parsevasp``.
    :type poscar: object
    :returns: A dictionary representation which are ready to be used when creating an
              AiiDA ``StructureData`` instance.
    :rtype: dict
    """

    # Fetch a dictionary containing the entries, make sure all coordinates are
    # cartesian
    poscar_dict = poscar.get_dict(direct=False)
    # Inverted dictionary with element names being the keys and numbers being the values
    symbols = {value['symbol']: key for key, value in elements.items()}

    for site in poscar_dict['sites']:
        specie = site['specie']
        # User can specify whatever they want for the elements, but
        # the symbols entries in AiiDA only support the entries defined
        # in aiida.common.constants.elements{}

        # Strip trailing _ in case user specifies potential
        symbol = specie.split('_')[0].capitalize()
        # Check if leading entry is part of
        # aiida.common.constants.elements{}, otherwise set to X
        if symbol not in symbols:
            symbol = 'X'

        site['symbol'] = symbol
        site['kind_name'] = specie

    return poscar_dict



"""Test the standard stream parser."""
# pylint: disable=unused-import,redefined-outer-name,unused-argument,unused-wildcard-import,wildcard-import

import re

import pytest


@pytest.mark.parametrize(['stream_parser'], [('stdout/out',)], indirect=True)
def test_stream_parser(stream_parser):
    """Test that the stream parser works."""
    errors = stream_parser.errors
    assert errors
    assert errors[0].kind == 'ERROR'
    assert errors[0].regex == re.compile('internal error in subroutine IBZKPT')
    assert len(stream_parser.warnings) == 0
